"""
Test functions for the eqcorrscan.utils.trigger module
"""
import unittest


class TestTriggerMethods(unittest.TestCase):
    def test_parameter_read_write(self):
        """Test parameter writing and reading."""
        from eqcorrscan.utils.trigger import read_trigger_parameters
        from eqcorrscan.utils.trigger import TriggerParameters
        import os

        test_parameters = TriggerParameters()
        test_parameters.write('test_parameters.par')
        check_parameters = read_trigger_parameters('test_parameters.par')
        self.assertEqual(test_parameters, check_parameters[0])
        os.remove('test_parameters.par')

    def test_channel_loop(self):
        """Test trigger generation in internal loop."""
        import numpy as np
        from eqcorrscan.utils.trigger import _channel_loop
        from eqcorrscan.utils.trigger import TriggerParameters
        from obspy import Trace

        parameters = [TriggerParameters({'station': 'TEST',
                                         'channel': 'SHZ',
                                         'sta_len': 0.3,
                                         'lta_len': 10.0,
                                         'thr_on': 10,
                                         'thr_off': 3,
                                         'lowcut': 2,
                                         'highcut': 20})]
        tr = Trace()
        tr.data = np.random.randn(2000)
        tr.data[1000:1010] = [100, -80, 70, -65, 60, -52, 45, -30, 15, 5]
        tr.stats.sampling_rate = 100
        tr.stats.station = parameters[0]['station']
        tr.stats.channel = parameters[0]['channel']
        # Test without despike
        triggers = _channel_loop(tr=tr, parameters=parameters,
                                 max_trigger_length=100,
                                 despike=False)
        self.assertEqual(len(triggers), 1)
        # Test with despike
        triggers = _channel_loop(tr=tr, parameters=parameters,
                                 max_trigger_length=100,
                                 despike=True)
        self.assertEqual(len(triggers), 1)
        # Test with no filter
        parameters[0]['lowcut'] = None
        parameters[0]['highcut'] = None
        triggers = _channel_loop(tr=tr, parameters=parameters,
                                 max_trigger_length=100,
                                 despike=False)
        self.assertEqual(len(triggers), 1)
        # Test with lowpass
        parameters[0]['highcut'] = 20
        triggers = _channel_loop(tr=tr, parameters=parameters,
                                 max_trigger_length=100,
                                 despike=False)
        self.assertEqual(len(triggers), 1)
        # Test with highpass
        parameters[0]['highcut'] = None
        parameters[0]['lowcut'] = 2
        triggers = _channel_loop(tr=tr, parameters=parameters,
                                 max_trigger_length=100,
                                 despike=False)
        self.assertEqual(len(triggers), 1)

    def test_main_trigger_routine(self):
        """Test the network_trigger function."""
        from eqcorrscan.utils.trigger import network_trigger, TriggerParameters
        from obspy import read
        import os
        testing_path = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                                    'test_data', 'WAV', 'TEST_')
        st = read(os.path.join(testing_path,
                               '2013-09-01-0410-36.DFDPC_027_00'))
        parameters = []
        for tr in st:
            parameters.append(TriggerParameters({'station': tr.stats.station,
                                                 'channel': tr.stats.channel,
                                                 'sta_len': 0.5,
                                                 'lta_len': 10.0,
                                                 'thr_on': 10.0,
                                                 'thr_off': 3.0,
                                                 'lowcut': 2.0,
                                                 'highcut': 15.0}))
        triggers = network_trigger(st=st, parameters=parameters,
                                   thr_coincidence_sum=5, moveout=30,
                                   max_trigger_length=60, despike=False)
        self.assertEqual(len(triggers), 1)
        triggers = network_trigger(st=st, parameters=parameters,
                                   thr_coincidence_sum=5, moveout=30,
                                   max_trigger_length=60, despike=False,
                                   parallel=False)
        self.assertEqual(len(triggers), 1)


if __name__ == '__main__':
    unittest.main()



"""
Simple tutorial detailing how to generate a series of templates from catalog\
data available online.
"""

import logging

from obspy.clients.fdsn import Client
from obspy.core.event import Catalog

from eqcorrscan.utils.catalog_utils import filter_picks
from eqcorrscan.core import template_gen

# Set up logging
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s\t%(name)s\t%(levelname)s\t%(message)s")


def mktemplates(
        network_code='GEONET', plot=True, publicIDs=None):
    """Functional wrapper to make templates"""
    publicIDs = publicIDs or [
        '2016p008122', '2016p008353', '2016p008155', '2016p008194']
    client = Client(network_code)
    # We want to download a few events from an earthquake sequence, these are
    # identified by publiID numbers, given as arguments

    catalog = Catalog()
    for publicID in publicIDs:
        try:
            catalog += client.get_events(
                eventid=publicID, includearrivals=True)
        except TypeError:
            # Cope with some FDSN services not implementing includearrivals
            catalog += client.get_events(eventid=publicID)

    # Lets plot the catalog to see what we have
    if plot:
        catalog.plot(projection='local', resolution='h')

    # We don't need all the picks, lets take the information from the
    # five most used stations - note that this is done to reduce computational
    # costs.
    catalog = filter_picks(catalog, top_n_picks=5)
    # We only want the P picks in this example, but you can use others or all
    #  picks if you want.
    for event in catalog:
        for pick in event.picks:
            if pick.phase_hint == 'S':
                event.picks.remove(pick)

    # Now we can generate the templates
    templates = template_gen.template_gen(
        method='from_client', catalog=catalog, client_id=network_code,
        lowcut=2.0, highcut=9.0, samp_rate=20.0, filt_order=4, length=3.0,
        prepick=0.15, swin='all', process_len=3600, plot=plot)

    # We now have a series of templates! Using Obspy's Stream.write() method we
    # can save these to disk for later use.  We will do that now for use in the
    # following tutorials.
    for i, template in enumerate(templates):
        template.write('tutorial_template_' + str(i) + '.ms', format='MSEED')
        # Note that this will warn you about data types.  As we don't care
        # at the moment, whatever obspy chooses is fine.
    return


if __name__ == '__main__':
    """Wrapper for template creation"""
    import sys
    import warnings
    if not len(sys.argv) > 1:
        warnings.warn('Needs a network ID followed by a list of event IDs, ' +
                      'will run the test case instead')
        mktemplates()
    else:
        net_code = sys.argv[1]
        idlist = list(sys.argv)[2:]
        print(idlist)
        mktemplates(net_code, idlist)



"""
Tests for the eqcorrscan.utils.picker functions.
"""
import unittest
import os

from obspy import read

from eqcorrscan.utils.picker import stalta_pick, cross_net


class PickerTests(unittest.TestCase):
    def test_stalta_picker(self):
        """
        Tests the stalta picker.
        """
        st = read()
        event = stalta_pick(stream=st, stalen=0.2, ltalen=3, trig_on=6,
                            trig_off=1.2)
        self.assertEqual(len(event.picks), 2)
        # Test a lower threshold
        event = stalta_pick(stream=st, stalen=0.2, ltalen=3, trig_on=3,
                            trig_off=1.2)
        self.assertEqual(len(event.picks), 11)
        # Test a higher threshold
        event = stalta_pick(stream=st, stalen=0.2, ltalen=3, trig_on=10,
                            trig_off=1.2)
        self.assertEqual(len(event.picks), 0)

    def test_cross_net(self):
        """
        Test the WECC style picker
        """
        st = read(os.path.join(os.path.abspath(os.path.dirname(__file__)),
                               'test_data', 'WAV', 'TEST_',
                               '2013-09-01-0410-35.DFDPC_024_00'))
        event = cross_net(stream=st)
        self.assertEqual(len(event.picks), 24)
        event = cross_net(stream=st, env=False)
        self.assertEqual(len(event.picks), 24)


if __name__ == "__main__":
    unittest.main()



"""
A series of test functions for the core functions in EQcorrscan.
"""
import unittest
import numpy as np
import logging

from obspy import read, UTCDateTime

from eqcorrscan.utils.synth_seis import generate_synth_data
from eqcorrscan.utils.correlate import get_stream_xcorr
from eqcorrscan.core import lag_calc as lag_calc_module
from eqcorrscan.core.lag_calc import (
    _xcorr_interp, LagCalcError, _prepare_data, lag_calc, xcorr_pick_family,
    _concatenate_and_correlate)
from eqcorrscan.core.match_filter import Detection, Family, Party, Template
from eqcorrscan.helpers.mock_logger import MockLoggingHandler

np.random.seed(999)


class SyntheticTests(unittest.TestCase):
    """ Test lag-calc with synthetic data. """
    @classmethod
    def setUpClass(cls):
        samp_rate = 50
        cls.t_length = .75
        # Make some synthetic templates
        templates, data, seeds = generate_synth_data(
            nsta=5, ntemplates=5, nseeds=10, samp_rate=samp_rate,
            t_length=cls.t_length, max_amp=10, max_lag=15, phaseout="both",
            jitter=0, noise=False, same_phase=True)
        # Rename channels
        channel_mapper = {"SYN_Z": "HHZ", "SYN_H": "HHN"}
        for tr in data:
            tr.stats.channel = channel_mapper[tr.stats.channel]
        for template in templates:
            for tr in template:
                tr.stats.channel = channel_mapper[tr.stats.channel]
        cls.party = Party()
        t = 0
        data_start = data[0].stats.starttime
        for template, template_seeds in zip(templates, seeds):
            template_name = "template_{0}".format(t)
            detections = []
            for i, sample in enumerate(template_seeds["time"]):
                det = Detection(
                    template_name=template_name,
                    detect_time=data_start + (sample / samp_rate),
                    detect_val=template_seeds["SNR"][i] / len(data),
                    no_chans=len(data),
                    chans=[
                        (tr.stats.station, tr.stats.channel) for tr in data],
                    threshold=0.0, threshold_input=0.0, threshold_type="abs",
                    typeofdet="ccc")
                det._calculate_event(
                    template_st=template, estimate_origin=False)
                detections.append(det)
            # Make a fully formed Template
            _template = Template(
                name=template_name, st=template, lowcut=2.0, highcut=15.0,
                samp_rate=samp_rate, filt_order=4, process_length=86400,
                prepick=10. / samp_rate, event=None)
            family = Family(template=_template, detections=detections)
            cls.party += family
            t += 1
        cls.data = data

    def _prepare_data_checks(self, detect_stream_dict, shift_len, family):
        self.assertEqual(
            len(detect_stream_dict), len(family.detections))
        for detection_id, stream in detect_stream_dict.items():
            detection = [d for d in family if d.id == detection_id][0]
            self.assertEqual(len(stream), detection.no_chans)
            for tr in stream:
                self.assertAlmostEqual(
                    tr.stats.endtime - tr.stats.starttime,
                    (2 * shift_len) + self.t_length, 1)

    def test_prepare_data(self):
        shift_len = 0.2
        for family in self.party:
            detect_stream_dict = _prepare_data(
                family=family, detect_data=self.data, shift_len=shift_len)
            self._prepare_data_checks(detect_stream_dict=detect_stream_dict,
                                      family=family, shift_len=shift_len)

    def test_prepare_data_too_short(self):
        data = self.data.copy()
        data.trim(self.party[0][0].detect_time,
                  self.party[0][0].detect_time + 3)
        shift_len = 0.2
        detect_stream_dict = _prepare_data(
            family=self.party[0], detect_data=data, shift_len=shift_len)
        self.assertEqual(len(detect_stream_dict), 1)

    def test_prepare_data_duplicate_channels(self):
        data = self.data.copy()
        data += data[0].copy()
        with self.assertRaises(LagCalcError):
            _prepare_data(
                family=self.party[0], detect_data=data, shift_len=0.2)

    def test_prepare_data_masked(self):
        data = self.data.copy()
        data.cutout(self.party[0][0].detect_time,
                    self.party[0][0].detect_time + 3)
        data.merge()
        shift_len = 0.2
        detect_stream_dict = _prepare_data(
            family=self.party[0], detect_data=data, shift_len=shift_len)
        short_key = self.party[0][0].id
        for key, value in detect_stream_dict.items():
            detection = [d for d in self.party[0] if d.id == key][0]
            if key == short_key:
                self.assertNotEqual(len(value), detection.no_chans)
            else:
                self.assertEqual(len(value), detection.no_chans)

    def test_family_picking(self):
        catalog_dict = xcorr_pick_family(
            family=self.party[0], stream=self.data, shift_len=0.2, plot=False)
        for event in catalog_dict.values():
            self.assertEqual(len(event.picks), len(self.data))
            for pick in event.picks:
                self.assertTrue("cc_max=" in pick.comments[0].text)
                self.assertAlmostEqual(
                    float(pick.comments[0].text.split("=")[-1]), 1.0, 1)

    def test_family_picking_missing_data(self):
        """ Check that this all works okay when one channel has some gaps """
        gappy_data = self.data.copy()
        sr = gappy_data[0].stats.sampling_rate
        mask = np.zeros(len(gappy_data[0].data), dtype=bool)
        gap_start = self.party[0][5].detect_time - 5
        gap_length = 3600
        gap_end = gap_start + gap_length
        gap_start = int(sr * (gap_start - UTCDateTime(0)))
        gap_end = int(sr * (gap_end - UTCDateTime(0)))
        mask[gap_start: gap_end] = np.ones(int(gap_length * sr), dtype=bool)
        gappy_data[0].data = np.ma.masked_array(
            data=gappy_data[0].data, mask=mask)
        catalog_dict = xcorr_pick_family(
            family=self.party[0], stream=gappy_data, shift_len=0.2, plot=False)
        gap = gappy_data.split().get_gaps()
        for event in catalog_dict.values():
            if len(event.picks) != len(self.data):
                self.assertEqual(len(event.picks), len(self.data) - 1)
                # Check that the event happened to be in the gap
                self.assertTrue(gap[0][4] <= event.picks[0].time <= gap[0][5])
                # Check that there isn't a pick on the channel missing data
                self.assertNotIn('.'.join(gap[0][0:4]),
                                 [pick.waveform_id.get_seed_string()
                                  for pick in event.picks])
            for pick in event.picks:
                self.assertTrue("cc_max=" in pick.comments[0].text)
                self.assertAlmostEqual(
                    float(pick.comments[0].text.split("=")[-1]), 1.0, 1)

    def test_family_picking_with_interpolation(self):
        catalog_dict = xcorr_pick_family(
            family=self.party[0], stream=self.data, shift_len=0.2, plot=False,
            interpolate=True)
        for event in catalog_dict.values():
            for pick in event.picks:
                self.assertTrue("cc_max=" in pick.comments[0].text)
                self.assertAlmostEqual(
                    float(pick.comments[0].text.split("=")[-1]), 1.0, 1)

    def test_lag_calc_api(self):
        detections = [d for f in self.party for d in f]
        templates = [f.template.st for f in self.party]
        template_names = [f.template.name for f in self.party]
        output_cat = lag_calc(
            detections, self.data, template_names, templates,
            shift_len=0.2, min_cc=0.4, horizontal_chans=['E', 'N', '1', '2'],
            vertical_chans=['Z'], cores=1, interpolate=False,
            plot=False)
        self.assertEqual(len(output_cat), len(detections))
        for event in output_cat:
            self.assertEqual(len(event.picks), len(self.data))


class SimpleRealDataTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        st = read().detrend().filter("bandpass", freqmin=2, freqmax=20)
        cls.template = st.slice(starttime=st[0].stats.starttime + 4,
                                endtime=st[0].stats.starttime + 10).copy()
        cls.shift_len = 1.0
        cls.detect_streams = []
        for i in range(1, 8):
            cls.detect_streams.append(
                st.slice(starttime=st[0].stats.starttime + i - cls.shift_len,
                         endtime=st[0].stats.starttime + i + 6 + cls.shift_len
                         ).copy())

    def test_correlation_max_and_position(self):
        ccc, chans = _concatenate_and_correlate(
            streams=self.detect_streams, template=self.template, cores=1)
        samp_rate = self.template[0].stats.sampling_rate
        t_start, t_end = (
            self.template[0].stats.starttime, self.template[0].stats.endtime)
        for _ccc, detect_stream in zip(ccc, self.detect_streams):
            d_start, d_end = (detect_stream[0].stats.starttime,
                              detect_stream[0].stats.endtime)
            if d_start <= t_start and d_end >= t_end:
                for ccc_chan in _ccc:
                    self.assertEqual(ccc_chan.max(), 1.0)
                    self.assertEqual(ccc_chan.argmax(),
                                     samp_rate * (t_start - d_start))
            else:
                for ccc_chan in _ccc:
                    self.assertNotEqual(ccc_chan.max(), 1.0)

    def test_correlation_precision(self):
        """Compare to correlation function outputs"""
        ccc, chans = _concatenate_and_correlate(
            streams=self.detect_streams, template=self.template, cores=1)
        fftw_xcorr_func = get_stream_xcorr("fftw")
        for _ccc, detect_stream in zip(ccc, self.detect_streams):
            fftw_ccc, _, _ = fftw_xcorr_func(
                templates=[self.template], stream=detect_stream, stack=False)
            for chan_ccc, fftw_chan_ccc in zip(_ccc, fftw_ccc[0]):
                self.assertTrue(np.allclose(
                    chan_ccc, fftw_chan_ccc, atol=.00001))


class ShortTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        log = logging.getLogger(lag_calc_module.__name__)
        cls._log_handler = MockLoggingHandler(level='DEBUG')
        log.addHandler(cls._log_handler)
        cls.log_messages = cls._log_handler.messages

    def setUp(self):
        self._log_handler.reset()

    def test_error(self):
        with self.assertRaises(LagCalcError):
            raise LagCalcError('Generic error')
        err = LagCalcError('Generic error')
        self.assertEqual('Generic error', err.value)
        self.assertEqual('Generic error', err.__repr__())
        self.assertEqual('LagCalcError: Generic error', err.__str__())

    def test_bad_interp(self):
        ccc = np.array([-0.21483282, -0.59443731, 0.1898917, -0.67516038,
                        0.60129057, -0.71043723,  0.16709118, 0.96839009,
                        1.58283915, -0.3053663])

        _xcorr_interp(ccc, 0.1)
        self.assertEqual(len(self.log_messages['warning']), 1)
        self.assertTrue(
            'not give an accurate result' in self.log_messages['warning'][0])


if __name__ == '__main__':
    unittest.main()



"""
Functions to test the functions within the eqcorrscan.utils.catalog_to_dd.py \
submodule.  Uses test data distributed with the EQcorrscan package.
"""
import unittest
import os
import numpy as np

from collections import Counter, namedtuple
from http.client import IncompleteRead

from obspy import UTCDateTime, Stream
from obspy.clients.fdsn import Client
from obspy.geodetics import gps2dist_azimuth

from eqcorrscan.utils.catalog_to_dd import (
    write_catalog, write_correlations, read_phase, write_event, _DTObs,
    _EventPair, _generate_event_id_mapper, _make_sparse_event,
    _prepare_stream, compute_differential_times, _filter_stream,
    _hypodd_event_str, write_phase, write_station)


class TestHelperObjects(unittest.TestCase):
    def test_dtobs(self):
        dtobs = _DTObs(station="FOZ", tt1=3.268, tt2=1.2857650,
                       weight=0.873265, phase="P")
        self.assertEqual(dtobs.ct_string, "FOZ       3.268   1.286 0.8733 P")
        self.assertEqual(dtobs.cc_string, "FOZ       1.982 0.8733 P")

    def test_event_pair(self):
        event_pair = _EventPair(event_id_1=12, event_id_2=54)
        event_pair.obs = [
            _DTObs(station="FOZ", tt1=3.268, tt2=1.2857650,
                   weight=0.873265, phase="P"),
            _DTObs(station="GCSZ", tt1=0.263, tt2=1.50,
                   weight=1.0, phase="S")]
        self.assertEqual(
            event_pair.ct_string,
            '#        12        54\nFOZ       3.268   1.286 0.8733 P\n'
            'GCSZ      0.263   1.500 1.0000 S')
        self.assertEqual(
            event_pair.cc_string,
            '#        12        54 0.0\nFOZ       1.982 0.8733 P\n'
            'GCSZ     -1.237 1.0000 S')


class TestCatalogMethods(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        starttime = UTCDateTime(2019, 8, 12, 10)
        endtime = UTCDateTime(2019, 8, 13)
        client = Client("GEONET")
        catalog = client.get_events(
            starttime=starttime, endtime=endtime, latitude=-44.5,
            longitude=167.9, maxradius=0.2)
        catalog.events.sort(
            key=lambda e: (e.preferred_origin() or e.origins[0]).time)
        StationInfo = namedtuple(
            "StationInfo", ["network", "station", "location"])
        picked_stations = [StationInfo(p.waveform_id.network_code,
                                       p.waveform_id.station_code,
                                       p.waveform_id.location_code)
                           for ev in catalog for p in ev.picks]
        stations_to_download = [
            sta for sta, _ in Counter(picked_stations).most_common(5)]
        streams = []
        for event in catalog[0:10]:  # Just get the first 10 events
            print(event.preferred_origin().time)
            stream = Stream()
            for station in stations_to_download:
                print(station)
                try:
                    stream += client.get_waveforms(
                        network=station.network, station=station.station,
                        location=station.location, channel="HH?",
                        starttime=event.preferred_origin().time - 10,
                        endtime=event.preferred_origin().time + 80)
                except IncompleteRead:
                    print(f"Could not download {station.station}")
            streams.append(stream)

        picked_stations = set(picked_stations)
        inv_bulk = [(sta.network, sta.station, sta.location, "HH?",
                     starttime, endtime) for sta in picked_stations]
        inventory = client.get_stations_bulk(inv_bulk, level="station")

        cls.streams = streams
        cls.catalog = catalog
        cls.inventory = inventory

    def test_id_mapper(self):
        map_one = _generate_event_id_mapper(self.catalog)
        for event in self.catalog:
            self.assertIn(event.resource_id.id, map_one.keys())
        self.assertEqual(len(self.catalog), len(map_one))
        map_one.pop(self.catalog[10].resource_id.id)
        self.assertNotEqual(len(self.catalog), len(map_one))
        map_two = _generate_event_id_mapper(self.catalog, map_one)
        self.assertEqual(len(self.catalog), len(map_two))
        for event in self.catalog:
            self.assertIn(event.resource_id.id, map_two.keys())
        # Check that a new event mapping has been made.
        self.assertEqual(len(self.catalog) + 1,
                         map_two.pop(self.catalog[10].resource_id.id))

    def test_sparse_event(self):
        for event in self.catalog:
            sparse_event = _make_sparse_event(event)
            self.assertEqual(sparse_event.resource_id, event.resource_id.id)
            self.assertEqual(
                sparse_event.origin_time, event.preferred_origin().time)
            self.assertEqual(len(sparse_event.picks), len(event.picks))
            for pick in event.picks:
                matched_pick = [
                    p for p in sparse_event.picks
                    if p.seed_id == pick.waveform_id.get_seed_string()
                    and p.phase == pick.phase_hint]
                self.assertEqual(len(matched_pick), 1)
                self.assertEqual(
                    matched_pick[0].tt, pick.time - sparse_event.origin_time)

    def test_process_stream(self):
        stream = self.streams[0]
        seed_ids = {tr.id for tr in stream}
        event = self.catalog[0]
        extract_len = 10.
        sliced_stream = _prepare_stream(
            stream, event, extract_len=extract_len, pre_pick=1.2,
            seed_pick_ids=None)
        p_picks = [p for p in event.picks if p.phase_hint[0] == "P"
                   and p.waveform_id.get_seed_string() in seed_ids]
        s_picks = [p for p in event.picks if p.phase_hint[0] == "S"
                   and p.waveform_id.get_seed_string() in seed_ids]
        self.assertEqual(len(p_picks), len(sliced_stream["P"]))
        self.assertEqual(len(s_picks), len(sliced_stream["S"]))
        for stream in sliced_stream.values():
            for tr in stream:
                self.assertEqual(
                    tr.stats.endtime - tr.stats.starttime, extract_len)

    def test_read_phase(self):
        """Function to test the phase reading function"""
        test_file = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                                 'test_data', 'tunnel.phase')
        test_catalog = read_phase(test_file)
        self.assertEqual(len(test_catalog), 2)
        self.assertEqual(test_catalog[0].origins[0].latitude, -43.169)
        self.assertEqual(test_catalog[0].origins[0].longitude, 170.646)
        self.assertEqual(test_catalog[0].origins[0].depth, -288)
        self.assertEqual(test_catalog[0].origins[0].time,
                         UTCDateTime('2012-01-30T01:45:43.25'))
        self.assertEqual(test_catalog[1].origins[0].latitude, -43.169)
        self.assertEqual(test_catalog[1].origins[0].longitude, 170.646)
        self.assertEqual(test_catalog[1].origins[0].depth, -288)
        self.assertEqual(test_catalog[1].origins[0].time,
                         UTCDateTime('2012-01-30T06:48:43.07'))

    def test_compute_differential_times(self):
        max_sep = 8.
        diff_times, mapper = compute_differential_times(
            catalog=self.catalog, correlation=False, event_id_mapper=None,
            max_sep=max_sep, min_link=8)
        reverse_mapper = {value: key for key, value in mapper.items()}
        self.assertEqual(len(self.catalog), len(diff_times))
        for key, links in diff_times.items():
            master_id = mapper[key]
            master_event = [e for e in self.catalog
                            if e.resource_id.id == key][0]
            for link in links:
                self.assertEqual(master_id, link.event_id_1)
                linked_event = [
                    e for e in self.catalog
                    if e.resource_id.id == reverse_mapper[link.event_id_2]][0]
                dist, _, _ = gps2dist_azimuth(
                    lat1=master_event.preferred_origin().latitude,
                    lon1=master_event.preferred_origin().longitude,
                    lat2=linked_event.preferred_origin().latitude,
                    lon2=linked_event.preferred_origin().longitude)
                self.assertLess(dist / 1000, max_sep)

    def test_compute_correlation_times(self):
        shift_len = 2
        short_cat = self.catalog[0:10]
        stream_dict = {event.resource_id.id: stream
                       for event, stream in zip(short_cat, self.streams)}
        for interpolate in [True, False]:
            diff_times, mapper = compute_differential_times(
                catalog=short_cat, correlation=True, event_id_mapper=None,
                max_sep=8., min_link=0, min_cc=0.0, stream_dict=stream_dict,
                extract_len=2.0, pre_pick=0.5, shift_len=shift_len,
                interpolate=interpolate, include_master=True)
            diff_times_cat, _ = compute_differential_times(
                catalog=short_cat, correlation=False, event_id_mapper=mapper,
                include_master=True)
            self.assertEqual(len(diff_times), len(short_cat))
            for master_id, linked in diff_times.items():
                for link in linked:
                    cat_link = [pair for pair in diff_times_cat[master_id]
                                if pair.event_id_2 == link.event_id_2][0]
                    if link.event_id_2 == link.event_id_1:
                        # This is the event matched with itself, check that tt1
                        # and tt2 are the same.
                        for obs in link.obs:
                            self.assertTrue(
                                np.allclose(obs.tt1, obs.tt2, atol=0.000001))
                    for obs in link.obs:
                        cat_obs = [o for o in cat_link.obs
                                   if o.station == obs.station and
                                   o.phase == obs.phase][0]
                        self.assertEqual(obs.tt1, cat_obs.tt1)
                        self.assertLessEqual(
                            abs(obs.tt2 - cat_obs.tt2), shift_len)
                        self.assertLessEqual(obs.weight, 1.0)

    def test_write_catalog(self):
        # Contents checked elsewhere
        write_catalog(catalog=self.catalog, event_id_mapper=None,
                      max_sep=8., min_link=8)
        self.assertTrue(os.path.isfile("dt.ct"))
        os.remove("dt.ct")

    def test_write_correlations(self):
        # Contents checked elsewhere
        shift_len = 2
        short_cat = self.catalog[0:10]
        stream_dict = {event.resource_id.id: stream
                       for event, stream in zip(short_cat, self.streams)}
        write_correlations(
            catalog=short_cat, event_id_mapper=None,
            max_sep=8., min_link=0, min_cc=0.0, stream_dict=stream_dict,
            extract_len=2.0, pre_pick=0.5, shift_len=shift_len,
            interpolate=False)
        self.assertTrue(os.path.isfile("dt.cc"))
        os.remove('dt.cc')

    def test_filter_stream(self):
        """ Check that original data are unchanged. """
        st = self.streams[0].copy()
        event_id = self.catalog[0].resource_id.id
        freqmin = (2.0, None)
        freqmax = (10.0, None)
        for _freqmin in freqmin:
            for _freqmax in freqmax:
                st_out = _filter_stream(st=st, event_id=event_id,
                                        highcut=_freqmax, lowcut=_freqmin)
                if _freqmax or _freqmin:
                    self.assertNotEqual(st_out[event_id], st)
                else:
                    self.assertEqual(st_out[event_id], st)

    def test_write_phase(self):
        """ This file has been tested with ph2dt """
        write_phase(self.catalog)
        test_data_path = os.path.join(
            os.path.abspath(os.path.dirname(__file__)), 'test_data')
        with open(os.path.join(test_data_path, "phase.dat"), "r") as f:
            original_phase = f.read()
        with open("phase.dat") as f:
            phase = f.read()
        self.assertEqual(phase, original_phase)

    def test_write_station(self):
        """ This file has been tested with ph2dt """
        write_station(self.inventory)
        test_data_path = os.path.join(
            os.path.abspath(os.path.dirname(__file__)), 'test_data')
        with open(os.path.join(test_data_path, "station.dat"), "r") as f:
            original_station = f.read()
        with open("station.dat") as f:
            station = f.read()
        self.assertEqual(station, original_station)

    def test_write_event(self):
        # Contents checked below
        write_event(self.catalog)
        self.assertTrue(os.path.isfile("event.dat"))
        os.remove("event.dat")

    def test_event_string(self):
        event_str = _hypodd_event_str(event=self.catalog[0], event_id=1)
        self.assertEqual(event_str, '20190812  10352775  -44.5152   167.8586'
                                    '    12.0000   5.51   0.01   0.00   '
                                    '0.74          1')


if __name__ == '__main__':
    unittest.main()



"""Functions to find peaks in data above a certain threshold.

:copyright:
    EQcorrscan developers.

:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import ctypes
import random
import logging
import numpy as np

from multiprocessing import Pool, cpu_count
from scipy import ndimage
from future.utils import native_str

from eqcorrscan.utils.correlate import pool_boy
from eqcorrscan.utils.libnames import _load_cdll


Logger = logging.getLogger(__name__)


def is_prime(number):
    """
    Function to test primality of a number. Function lifted from online
    resource:
        http://www.codeproject.com/Articles/691200/Primality-test-algorithms-Prime-test-The-fastest-w

    This function is distributed under a separate licence:
        This article, along with any associated source code and files, is \
        licensed under The Code Project Open License (CPOL)

    :type number: int
    :param number: Integer to test for primality

    :returns: bool

    >>> is_prime(4)
    False
    >>> is_prime(3)
    True
    """
    ''' if number != 1 '''
    if number > 1:
        ''' repeat the test few times '''
        for time in range(3):
            ''' Draw a RANDOM number in range of number ( Z_number )  '''
            randomNumber = random.randint(2, number - 1)
            ''' Test if a^(n-1) = 1 mod n '''
            if pow(randomNumber, number - 1, number) != 1:
                return False
        return True
    else:
        ''' case number == 1 '''
        return False


def find_peaks_compiled(arr, thresh, trig_int, full_peaks=False):
    """
    Determine peaks in an array of data above a certain threshold.

    :type arr: numpy.ndarray
    :param arr: 1-D numpy array is required
    :type thresh: float
    :param thresh:
        The threshold below which will be considered noise and peaks will
        not be found in.
    :type trig_int: int
    :param trig_int:
        The minimum difference in samples between triggers, if multiple
        peaks within this window this code will find the highest.
    :type full_peaks: bool
    :param full_peaks:
        If True, will declustering within data-sections above the threshold,
        rather than just taking the peak within that section. This will take
        more time. This defaults to True for match_filter.

    :return: peaks: Lists of tuples of peak values and locations.
    :rtype: list
    """
    if not np.any(np.abs(arr) > thresh):
        # Fast fail
        return []
    if not full_peaks:
        peak_vals, peak_indices = _find_peaks_c(array=arr, threshold=thresh)
    else:
        peak_vals = arr
        peak_indices = np.arange(arr.shape[0])
    if len(peak_vals) > 0:
        peaks = decluster(
            peaks=np.array(peak_vals), index=np.array(peak_indices),
            trig_int=trig_int + 1, threshold=thresh)
        peaks = sorted(peaks, key=lambda peak: peak[1], reverse=False)
        return peaks
    else:
        return []


def find_peaks2_short(arr, thresh, trig_int, full_peaks=False):
    """
    Determine peaks in an array of data above a certain threshold.

    Uses a mask to remove data below threshold and finds peaks in what is left.

    :type arr: numpy.ndarray
    :param arr: 1-D numpy array is required
    :type thresh: float
    :param thresh:
        The threshold below which will be considered noise and peaks will
        not be found in.
    :type trig_int: int
    :param trig_int:
        The minimum difference in samples between triggers, if multiple
        peaks within this window this code will find the highest.
    :type full_peaks: bool
    :param full_peaks:
        If True, will remove the issue eluded to below, by declustering within
        data-sections above the threshold, rather than just taking the peak
        within that section. This will take more time. This defaults to True
        for match_filter.

    :return: peaks: Lists of tuples of peak values and locations.
    :rtype: list


    >>> import numpy as np
    >>> arr = np.random.randn(100)
    >>> threshold = 10
    >>> arr[40] = 20
    >>> arr[60] = 100
    >>> find_peaks2_short(arr, threshold, 3)
    [(20.0, 40), (100.0, 60)]
    """
    # Set everything below the threshold to zero
    image = np.copy(arr)
    Logger.debug("Threshold: {0}\tMax: {1}".format(thresh, max(image)))
    image[np.abs(image) < thresh] = 0
    if len(image[np.abs(image) > thresh]) == 0:
        Logger.debug("No values over threshold {0}".format(thresh))
        return []
    if np.all(np.abs(arr) > thresh):
        full_peaks = True
    Logger.debug('Found {0} samples above the threshold'.format(
        len(image[image > thresh])))
    initial_peaks = []
    # Find the peaks
    labeled_image, number_of_objects = ndimage.label(image)
    peak_slices = ndimage.find_objects(labeled_image)
    for peak_slice in peak_slices:
        window = arr[peak_slice[0].start: peak_slice[0].stop]
        if peak_slice[0].stop - peak_slice[0].start >= trig_int and full_peaks:
            window_peaks, window_peak_indexes = ([], [])
            for i in np.arange(peak_slice[0].start, peak_slice[0].stop):
                if i == peak_slice[0].start:
                    prev_value = 0
                else:
                    prev_value = arr[i - 1]
                if i == peak_slice[0].stop - 1:
                    next_value = 0
                else:
                    next_value = arr[i + 1]
                # Check for consistent sign - either both greater or
                # both smaller.
                if (next_value - arr[i]) * (prev_value - arr[i]) > 0:
                    window_peaks.append(arr[i])
                    window_peak_indexes.append(i)
            peaks = decluster(
                peaks=np.array(window_peaks), trig_int=trig_int + 1,
                index=np.array(window_peak_indexes))
        else:
            peaks = [(window[np.argmax(abs(window))],
                      int(peak_slice[0].start + np.argmax(abs(window))))]
        initial_peaks.extend(peaks)
    peaks = decluster(peaks=np.array(list(zip(*initial_peaks))[0]),
                      index=np.array(list(zip(*initial_peaks))[1]),
                      trig_int=trig_int + 1)
    if initial_peaks:
        peaks = sorted(peaks, key=lambda time: time[1], reverse=False)
        return peaks
    else:
        Logger.info('No peaks for you!')
        return []


def multi_find_peaks(arr, thresh, trig_int, parallel=True, full_peaks=False,
                     cores=None, internal_func=find_peaks_compiled):
    """
    Wrapper for find-peaks for multiple arrays.

    :type arr: numpy.ndarray
    :param arr: 2-D numpy array is required
    :type thresh: list
    :param thresh:
        The threshold below which will be considered noise and peaks will not
        be found in. One threshold per array.
    :type trig_int: int
    :param trig_int:
        The minimum difference in samples between triggers, if multiple
        peaks within this window this code will find the highest.
    :type parallel: bool
    :param parallel:
        Whether to compute in parallel or not - will use multiprocessing if not
        using the compiled internal_func
    :type full_peaks: bool
    :param full_peaks: See `eqcorrscan.utils.findpeaks.find_peaks2_short`
    :type cores: int
    :param cores:
        Maximum number of processes to spin up for parallel peak-finding
    :type internal_func: callable
    :param internal_func:
        Function to use for peak finding - defaults to the compiled version.

    :returns:
        List of list of tuples of (peak, index) in same order as input arrays
    """
    peaks = []
    if not parallel:
        for sub_arr, arr_thresh in zip(arr, thresh):
            peaks.append(internal_func(
                arr=sub_arr, thresh=arr_thresh, trig_int=trig_int,
                full_peaks=full_peaks))
    else:
        if cores is None:
            cores = min(arr.shape[0], cpu_count())
        if internal_func.__name__ != 'find_peaks_compiled':
            with pool_boy(Pool=Pool, traces=arr.shape[0], cores=cores) as pool:
                params = ((sub_arr, arr_thresh, trig_int, full_peaks)
                          for sub_arr, arr_thresh in zip(arr, thresh))
                results = [
                    pool.apply_async(internal_func, param) for param in params]
                peaks = [res.get() for res in results]
        else:
            peaks = _multi_find_peaks_compiled(
                arr, thresh, trig_int, full_peaks=full_peaks, cores=cores)
    return peaks


def _multi_find_peaks_compiled(arrays, thresholds, trig_int, full_peaks,
                               cores):
    """
    Determine peaks in an array or arrays of data above a certain threshold.

    :type arrays: numpy.ndarray
    :param arrays: 2-D numpy array is required
    :type thresholds: list
    :param thresholds:
        Minimum value for peaks.
    :type trig_int: int
    :param trig_int:
        The minimum difference in samples between triggers, if multiple
        peaks within this window this code will find the highest.
    :type full_peaks: bool
    :param full_peaks:
        If True, will decluster within data-sections above the threshold,
        rather than just taking the peak within that section. This will take
        more time. This defaults to True for match_filter.
    :type cores: int
    :param cores: Number of threads to parallel across

    :return: peaks: List of List of tuples of peak values and locations.
    :rtype: list
    """
    if not full_peaks:
        peak_vals, peak_indices = _multi_find_peaks_c(
            arrays=arrays, thresholds=thresholds, threads=cores)
        # Remove empty arrays
        peak_mapper = {}
        map_index = 0
        _peak_vals = []
        _peak_indices = []
        _thresholds = []
        for i in range(arrays.shape[0]):
            if len(peak_vals[i]) > 0:
                peak_mapper.update({i: map_index})
                _peak_vals.append(peak_vals[i])
                _peak_indices.append(peak_indices[i])
                _thresholds.append(thresholds[i])
                map_index += 1
        peak_vals = _peak_vals
        peak_indices = _peak_indices
        thresholds = _thresholds
    else:
        peak_vals = arrays
        peak_indices = [np.arange(arr.shape[0]) for arr in arrays]
        peak_mapper = {i: i for i in range(len(peak_indices))}
    if len(peak_indices) > 0:
        peaks = _multi_decluster(
            peaks=peak_vals, indices=peak_indices, trig_int=trig_int,
            thresholds=thresholds, cores=cores)
        peaks = [sorted(_peaks, key=lambda peak: peak[1], reverse=False)
                 for _peaks in peaks]
    out_peaks = []
    for i in range(arrays.shape[0]):
        if i in peak_mapper.keys():
            out_peaks.append(peaks[peak_mapper[i]])
        else:
            out_peaks.append([])
    return out_peaks


def _multi_decluster(peaks, indices, trig_int, thresholds, cores):
    """
    Decluster peaks based on an enforced minimum separation.

    Only works when peaks and indices are all the same shape.

    :type peaks: list
    :param peaks: list of arrays of peak values
    :type indices: list
    :param indices: list of arrays of locations of peaks
    :type trig_int: int
    :param trig_int: Minimum trigger interval in samples
    :type thresholds: list
    :param thresholds: list of float of threshold values

    :return: list of lists of tuples of (value, sample)
    """
    utilslib = _load_cdll('libutils')

    lengths = np.array([peak.shape[0] for peak in peaks], dtype=int)
    trig_int = int(trig_int)
    n = np.int32(len(peaks))
    cores = min(cores, n)

    total_length = lengths.sum()

    max_indexes = [_indices.max() for _indices in indices]
    max_index = max(max_indexes)
    for var in [trig_int, lengths.max(), max_index]:
        if var == ctypes.c_long(var).value:
            long_type = ctypes.c_long
            func = utilslib.multi_decluster
        elif var == ctypes.c_longlong(var).value:
            long_type = ctypes.c_longlong
            func = utilslib.multi_decluster_ll
        else:
            # Note, could use numpy.gcd to try and find greatest common
            # divisor and make numbers smaller
            raise OverflowError("Maximum index larger than internal long long")

    func.argtypes = [
        np.ctypeslib.ndpointer(dtype=np.float32, shape=(total_length,),
                               flags=native_str('C_CONTIGUOUS')),
        np.ctypeslib.ndpointer(dtype=long_type, shape=(total_length,),
                               flags=native_str('C_CONTIGUOUS')),
        np.ctypeslib.ndpointer(dtype=long_type, shape=(n,),
                               flags=native_str('C_CONTIGUOUS')),
        ctypes.c_int,
        np.ctypeslib.ndpointer(dtype=np.float32, shape=(n,),
                               flags=native_str('C_CONTIGUOUS')),
        long_type,
        np.ctypeslib.ndpointer(dtype=np.uint32, shape=(total_length,),
                               flags=native_str('C_CONTIGUOUS')),
        ctypes.c_int]
    func.restype = ctypes.c_int

    peaks_sorted = np.empty(total_length, dtype=np.float32)
    indices_sorted = np.empty_like(peaks_sorted, dtype=np.float32)

    # TODO: When doing full decluster from match-filter, all lengths will be
    # TODO: the same - would be more efficient to use numpy sort on 2D matrix
    start_ind = 0
    end_ind = 0
    for _peaks, _indices, length in zip(peaks, indices, lengths):
        end_ind += length
        sorted_indices = np.abs(_peaks).argsort()
        peaks_sorted[start_ind: end_ind] = _peaks[sorted_indices[::-1]]
        indices_sorted[start_ind: end_ind] = _indices[sorted_indices[::-1]]
        start_ind += length

    peaks_sorted = np.ascontiguousarray(peaks_sorted, dtype=np.float32)
    indices_sorted = np.ascontiguousarray(
        indices_sorted, dtype=long_type)
    lengths = np.ascontiguousarray(lengths, dtype=long_type)
    thresholds = np.ascontiguousarray(thresholds, dtype=np.float32)
    out = np.zeros(total_length, dtype=np.uint32)
    ret = func(
        peaks_sorted, indices_sorted, lengths, np.int32(n), thresholds,
        long_type(trig_int + 1), out, np.int32(cores))
    if ret != 0:
        raise MemoryError("Issue with c-routine, returned %i" % ret)

    peaks_out = []
    slice_start = 0
    for length in lengths:
        slice_end = slice_start + length
        out_mask = out[slice_start: slice_end].astype(bool)
        declustered_peaks = peaks_sorted[slice_start: slice_end][out_mask]
        declustered_indices = indices_sorted[slice_start: slice_end][out_mask]
        peaks_out.append(list(zip(declustered_peaks, declustered_indices)))
        slice_start = slice_end
    return peaks_out


def decluster(peaks, index, trig_int, threshold=0):
    """
    Decluster peaks based on an enforced minimum separation.

    :type peaks: np.array
    :param peaks: array of peak values
    :type index: np.ndarray
    :param index: locations of peaks
    :type trig_int: int
    :param trig_int: Minimum trigger interval in samples

    :return: list of tuples of (value, sample)
    """
    utilslib = _load_cdll('libutils')

    length = peaks.shape[0]
    trig_int = int(trig_int)

    for var in [index.max(), trig_int]:
        if var == ctypes.c_long(var).value:
            long_type = ctypes.c_long
            func = utilslib.decluster
        elif var == ctypes.c_longlong(var).value:
            long_type = ctypes.c_longlong
            func = utilslib.decluster_ll
        else:
            raise OverflowError("Maximum index larger than internal long long")

    func.argtypes = [
        np.ctypeslib.ndpointer(dtype=np.float32, shape=(length,),
                               flags=native_str('C_CONTIGUOUS')),
        np.ctypeslib.ndpointer(dtype=long_type, shape=(length,),
                               flags=native_str('C_CONTIGUOUS')),
        long_type, ctypes.c_float, long_type,
        np.ctypeslib.ndpointer(dtype=np.uint32, shape=(length,),
                               flags=native_str('C_CONTIGUOUS'))]
    func.restype = ctypes.c_int

    sorted_inds = np.abs(peaks).argsort()
    arr = peaks[sorted_inds[::-1]]
    inds = index[sorted_inds[::-1]]
    arr = np.ascontiguousarray(arr, dtype=np.float32)
    inds = np.ascontiguousarray(inds, dtype=long_type)
    out = np.zeros(len(arr), dtype=np.uint32)

    ret = func(
        arr, inds, long_type(length), np.float32(threshold),
        long_type(trig_int), out)
    if ret != 0:
        raise MemoryError("Issue with c-routine, returned %i" % ret)

    peaks_out = list(zip(arr[out.astype(bool)], inds[out.astype(bool)]))
    return peaks_out


def _find_peaks_c(array, threshold):
    """
    Use a C func to find peaks in the array.
    """
    utilslib = _load_cdll('libutils')

    length = array.shape[0]
    utilslib.find_peaks.argtypes = [
        np.ctypeslib.ndpointer(dtype=np.float32, shape=(length, ),
                               flags=native_str('C_CONTIGUOUS')),
        ctypes.c_long, ctypes.c_float,
        np.ctypeslib.ndpointer(dtype=np.uint32, shape=(length, ),
                               flags=native_str('C_CONTIGUOUS'))]
    utilslib.find_peaks.restype = ctypes.c_int
    arr = np.ascontiguousarray(array, np.float32)
    out = np.ascontiguousarray(np.zeros((length, ), dtype=np.uint32))
    ret = utilslib.find_peaks(arr, ctypes.c_long(length), threshold, out)

    if ret != 0:
        raise MemoryError("Internal error")

    peaks_locations = np.nonzero(out)
    return array[peaks_locations], peaks_locations[0]


def _multi_find_peaks_c(arrays, thresholds, threads):
    """
    Wrapper for multi-find peaks C-func
    """
    utilslib = _load_cdll('libutils')

    length = arrays.shape[1]
    n = np.int32(arrays.shape[0])
    thresholds = np.ascontiguousarray(thresholds, np.float32)
    arr = np.ascontiguousarray(arrays.flatten(), np.float32)
    utilslib.multi_find_peaks.argtypes = [
        np.ctypeslib.ndpointer(dtype=np.float32, shape=(n * length,),
                               flags=native_str('C_CONTIGUOUS')),
        ctypes.c_long, ctypes.c_int,
        np.ctypeslib.ndpointer(dtype=np.float32, shape=(n, ),
                               flags=native_str('C_CONTIGUOUS')),
        ctypes.c_int,
        np.ctypeslib.ndpointer(dtype=np.uint32, shape=(n * length, ),
                               flags=native_str('C_CONTIGUOUS'))]
    utilslib.multi_find_peaks.restype = ctypes.c_int

    out = np.ascontiguousarray(np.zeros((n * length, ), dtype=np.uint32))
    ret = utilslib.multi_find_peaks(
        arr, ctypes.c_long(length), n, thresholds, threads, out)
    # Copy data to avoid farking the users data
    if ret != 0:
        raise MemoryError("Internal error")
    peaks = []
    peak_locations = []
    out = out.reshape(n, length)
    for i in range(n):
        peak_locs = np.nonzero(out[i])
        peaks.append(arrays[i][peak_locs])
        peak_locations.append(peak_locs[0])

    return peaks, peak_locations


def coin_trig(peaks, stachans, samp_rate, moveout, min_trig, trig_int):
    """
    Find network coincidence triggers within peaks of detection statistics.

    Useful for finding network detections from sets of detections on individual
    stations.

    :type peaks: list
    :param peaks: List of lists of tuples of (peak, index) for each \
        station-channel.  Index should be in samples.
    :type stachans: list
    :param stachans: List of tuples of (station, channel) in the order of \
        peaks.
    :type samp_rate: float
    :param samp_rate: Sampling rate in Hz
    :type moveout: float
    :param moveout: Allowable network moveout in seconds.
    :type min_trig: int
    :param min_trig: Minimum station-channels required to declare a trigger.
    :type trig_int: float
    :param trig_int:
        Minimum allowable time between network triggers in seconds.

    :return:
        List of tuples of (peak, index), for the earliest detected station.
    :rtype: list

    >>> peaks = [[(0.5, 100), (0.3, 800)], [(0.4, 120), (0.7, 850)]]
    >>> triggers = coin_trig(peaks, [('a', 'Z'), ('b', 'Z')], 10, 3, 2, 1)
    >>> print(triggers)
    [(0.45, 100)]
    """
    triggers = []
    for stachan, _peaks in zip(stachans, peaks):
        for peak in _peaks:
            trigger = (peak[1], peak[0], '.'.join(stachan))
            triggers.append(trigger)
    coincidence_triggers = []
    for i, master in enumerate(triggers):
        slaves = triggers[i + 1:]
        coincidence = 1
        trig_time = master[0]
        trig_val = master[1]
        for slave in slaves:
            if abs(slave[0] - master[0]) <= (moveout * samp_rate) and \
               slave[2] != master[2]:
                coincidence += 1
                if slave[0] < master[0]:
                    trig_time = slave[0]
                trig_val += slave[1]
        if coincidence >= min_trig:
            coincidence_triggers.append((trig_val / coincidence,
                                         trig_time))
    # Sort by trigger-value, largest to smallest - remove duplicate detections
    if coincidence_triggers:
        coincidence_triggers.sort(key=lambda tup: tup[0], reverse=True)
        output = [coincidence_triggers[0]]
        for coincidence_trigger in coincidence_triggers[1:]:
            add = True
            for peak in output:
                # If the event occurs within the trig_int time then do not add
                # it, and break out of the inner loop.
                if abs(coincidence_trigger[1] - peak[1]) < (trig_int *
                                                            samp_rate):
                    add = False
                    break
            if add:
                output.append((coincidence_trigger[0],
                               coincidence_trigger[1]))
        output.sort(key=lambda tup: tup[1])
        return output
    else:
        return []


if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""
This module contains functions relevant to executing subspace detection
for earthquake catalogs.

We recommend that you read Harris' detailed report on subspace detection
theory which can be found here: https://e-reports-ext.llnl.gov/pdf/335299.pdf

:copyright:
    EQcorrscan developers.

:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import numpy as np
import logging
import time
import h5py
import getpass
import eqcorrscan
import copy
import scipy

from obspy import Trace, UTCDateTime, Stream
from obspy.core.event import (
    Event, CreationInfo, ResourceIdentifier, Comment, WaveformStreamID, Pick)

from eqcorrscan.utils.clustering import svd
from eqcorrscan.utils import findpeaks, pre_processing, stacking, plotting
from eqcorrscan.core.match_filter import Detection
from eqcorrscan.core.match_filter.helpers import extract_from_stream
from eqcorrscan.utils.plotting import subspace_detector_plot, subspace_fc_plot


Logger = logging.getLogger(__name__)


class Detector(object):
    """
    Class to serve as the base for subspace detections.

    :type name: str
    :param name: Name of subspace detector, used for book-keeping
    :type sampling_rate: float
    :param sampling_rate: Sampling rate in Hz of original waveforms
    :type multiplex: bool
    :param multiplex: Is this detector multiplexed.
    :type stachans: list
    :param stachans:
        List of tuples of (station, channel) used in detector.
        If multiplexed, these must be in the order that multiplexing was done.
    :type lowcut: float
    :param lowcut: Lowcut filter in Hz
    :type highcut: float
    :param highcut: Highcut filter in Hz
    :type filt_order: int
    :param filt_order: Number of corners for filtering
    :type data: numpy.ndarray
    :param data: The actual detector
    :type u: numpy.ndarray
    :param u: Full rank U matrix of left (input) singular vectors.
    :type sigma: numpy.ndarray
    :param sigma: Full rank vector of singular values.
    :type v: numpy.ndarray
    :param v: Full rank right (output) singular vectors.
    :type dimension: int
    :param dimension: Dimension of data.

    .. warning::
        Changing between scipy.linalg.svd solvers (obvious changes between
        scipy version 0.18.x and 0.19.0) result in sign changes in svd results.
        You should only run a detector created using the same scipy version
        as you currently run.

    """
    def __init__(self, name=None, sampling_rate=None, multiplex=None,
                 stachans=None, lowcut=None, highcut=None,
                 filt_order=None, data=None, u=None, sigma=None, v=None,
                 dimension=None):
        self.name = name
        self.sampling_rate = sampling_rate
        self.multiplex = multiplex
        self.stachans = stachans
        # self.delays = delays
        self.lowcut = lowcut
        self.highcut = highcut
        self.filt_order = filt_order
        self.data = data
        self.u = u
        self.sigma = sigma
        self.v = v
        self.dimension = dimension

    def __repr__(self):
        if self.name:
            out = 'Detector: ' + self.name
        else:
            out = 'Empty Detector object'
        return out

    def __str__(self):
        out = 'Detector object: \n'
        for key in ['name', 'sampling_rate', 'multiplex', 'lowcut', 'highcut',
                    'filt_order', 'dimension']:
            if self.__getattribute__(key):
                out += ('\t' + key + ': ' + str(self.__getattribute__(key)) +
                        '\n')
        return out

    def __eq__(self, other):
        if not isinstance(other, Detector):
            return False
        for key in ['name', 'sampling_rate', 'multiplex', 'lowcut', 'highcut',
                    'filt_order', 'dimension', 'stachans']:
            if not self.__getattribute__(key) == other.__getattribute__(key):
                return False
        for key in ['data', 'u', 'v', 'sigma']:
            list_item = self.__getattribute__(key)
            other_list = other.__getattribute__(key)
            if not len(list_item) == len(other_list):
                return False
            for item, other_item in zip(list_item, other_list):
                if not np.allclose(np.absolute(item),
                                   np.absolute(other_item), atol=0.001):
                    return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __len__(self):
        return len(self.data)

    def construct(self, streams, lowcut, highcut, filt_order,
                  sampling_rate, multiplex, name, align, shift_len=0,
                  reject=0.3, no_missed=True, plot=False):
        """
        Construct a subspace detector from a list of streams, full rank.

        Subspace detector will be full-rank, further functions can be used \
        to select the desired dimensions.

        :type streams: list
        :param streams:
            List of :class:`obspy.core.stream.Stream` to be used to generate
            the subspace detector.  These should be pre-clustered and aligned.
        :type lowcut: float
        :param lowcut: Lowcut in Hz, can be None to not apply filter
        :type highcut: float
        :param highcut: Highcut in Hz, can be None to not apply filter
        :type filt_order: int
        :param filt_order: Number of corners for filter.
        :type sampling_rate: float
        :param sampling_rate: Desired sampling rate in Hz
        :type multiplex: bool
        :param multiplex:
            Whether to multiplex the data or not.  Data are multiplexed
            according to the method of Harris, see the multi function for
            details.
        :type name: str
        :param name: Name of the detector, used for book-keeping.
        :type align: bool
        :param align:
            Whether to align the data or not - needs to be done at some point
        :type shift_len: float
        :param shift_len: Maximum shift allowed for alignment in seconds.
        :type reject: float
        :param reject:
            Minimum correlation to include traces - only used if align=True.
        :type no_missed: bool
        :param no_missed:
            Reject streams with missed traces, defaults to True. A missing
            trace from lots of events will reduce the quality of the subspace
            detector if multiplexed.  Only used when multi is set to True.
        :type plot: bool
        :param plot: Whether to plot the alignment stage or not.

        .. note::
            The detector will be normalized such that the data, before
            computing the singular-value decomposition, will have unit energy.
            e.g. We divide the amplitudes of the data by the L1 norm of the
            data.

        .. warning::
            EQcorrscan's alignment will attempt to align over the whole data
            window given.  For long (more than 2s) chunks of data this can give
            poor results and you might be better off using the
            :func:`eqcorrscan.utils.stacking.align_traces` function externally,
            focusing on a smaller window of data.  To do this you would align
            the data prior to running construct.
        """
        self.lowcut = lowcut
        self.highcut = highcut
        self.filt_order = filt_order
        self.sampling_rate = sampling_rate
        self.name = name
        self.multiplex = multiplex
        # Pre-process data
        p_streams, stachans = _subspace_process(
            streams=copy.deepcopy(streams), lowcut=lowcut, highcut=highcut,
            filt_order=filt_order, sampling_rate=sampling_rate,
            multiplex=multiplex, align=align, shift_len=shift_len,
            reject=reject, plot=plot, no_missed=no_missed)
        # Compute the SVD, use the cluster.SVD function
        u, sigma, v, svd_stachans = svd(stream_list=p_streams, full=True)
        self.stachans = stachans
        # self.delays = delays
        self.u = u
        self.v = v
        self.sigma = sigma
        self.data = copy.deepcopy(u)  # Set the data matrix to be full rank U.
        self.dimension = np.inf
        return self

    def partition(self, dimension):
        """
        Partition subspace into desired dimension.

        :type dimension: int
        :param dimension: Maximum dimension to use.
        """
        # Take leftmost 'dimension' input basis vectors
        for i, channel in enumerate(self.u):
            if self.v[i].shape[1] < dimension:
                raise IndexError('Channel is max dimension %s'
                                 % self.v[i].shape[1])
            self.data[i] = channel[:, 0:dimension]
        self.dimension = dimension
        return self

    def energy_capture(self, stachans='all', size=(10, 7), show=False):
        """
        Calculate the average percentage energy capture for this subspace.

        :return: Percentage energy capture
        :rtype: float
        """
        if show:
            return subspace_fc_plot(detector=self, stachans=stachans,
                                    size=size, show=show)
        percent_capture = 0
        if np.isinf(self.dimension):
            return 100
        for channel in self.sigma:
            fc = np.sum(channel[0:self.dimension]) / np.sum(channel)
            percent_capture += fc
        else:
            return 100 * (percent_capture / len(self.sigma))

    def detect(self, st, threshold, trig_int, moveout=0, min_trig=0,
               process=True, extract_detections=False, cores=1):
        """
        Detect within continuous data using the subspace method.

        :type st: obspy.core.stream.Stream
        :param st:
            Un-processed stream to detect within using the subspace detector.
        :type threshold: float
        :param threshold: Threshold value for detections between 0-1
        :type trig_int: float
        :param trig_int: Minimum trigger interval in seconds.
        :type moveout: float
        :param moveout:
            Maximum allowable moveout window for non-multiplexed, network
            detection.  See note.
        :type min_trig: int
        :param min_trig:
            Minimum number of stations exceeding threshold for non-multiplexed,
            network detection. See note.
        :type process: bool
        :param process:
            Whether or not to process the stream according to the parameters
            defined by the detector.  Default is True, which will process the
            data.
        :type extract_detections: bool
        :param extract_detections:
            Whether to extract waveforms for each detection or not, if True
            will return detections and streams.
        :type cores: int
        :param cores: Number of threads to process data with.

        :return: list of :class:`eqcorrscan.core.match_filter.Detection`
        :rtype: list

        .. warning::
            Subspace is currently in beta, see note in the subspace tutorial
            for information.

        .. note::
            If running in bulk with detectors that all have the same
            parameters then you can pre-process the data and set process to
            False.  This will speed up this detect function dramatically.

        .. warning::
            If the detector and stream are multiplexed then they must
            contain the same channels and multiplexed in the same order. This
            is handled internally when process=True, but if running in bulk
            you must take care.

        .. note::
            Non-multiplexed, network detection.  When the detector is
            not multiplexed, but there are multiple channels within the
            detector, we do not stack the single-channel detection statistics
            because we do not have a one-size-fits-all solution for computing
            delays for a subspace detector (if you want to implement one, then
            please contribute it!).  Therefore, these parameters provide a
            means for declaring a network coincidence trigger using
            single-channel detection statistics, in a similar fashion to the
            commonly used network-coincidence trigger with energy detection
            statistics.
        """
        return _detect(detector=self, st=st, threshold=threshold,
                       trig_int=trig_int, moveout=moveout, min_trig=min_trig,
                       process=process, extract_detections=extract_detections,
                       cores=cores)

    def write(self, filename):
        """
        Write detector to a file - uses HDF5 file format.

        Meta-data are stored alongside numpy data arrays. See h5py.org for \
        details of the methods.

        :type filename: str
        :param filename: Filename to save the detector to.
        """
        f = h5py.File(filename, "w")
        # Must store eqcorrscan version number, username would be useful too.
        data_group = f.create_group(name="data")
        for i, data in enumerate(self.data):
            dset = data_group.create_dataset(name="data_" + str(i),
                                             shape=data.shape,
                                             dtype=data.dtype)
            dset[...] = data
        data_group.attrs['length'] = len(self.data)
        data_group.attrs['name'] = self.name.encode("ascii", "ignore")
        data_group.attrs['sampling_rate'] = self.sampling_rate
        data_group.attrs['multiplex'] = self.multiplex
        data_group.attrs['lowcut'] = self.lowcut
        data_group.attrs['highcut'] = self.highcut
        data_group.attrs['filt_order'] = self.filt_order
        data_group.attrs['dimension'] = self.dimension
        data_group.attrs['user'] = getpass.getuser()
        data_group.attrs['eqcorrscan_version'] = str(eqcorrscan.__version__)
        # Convert station-channel list to something writable
        ascii_stachans = ['.'.join(stachan).encode("ascii", "ignore")
                          for stachan in self.stachans]
        stachans = f.create_dataset(name="stachans",
                                    shape=(len(ascii_stachans),),
                                    dtype='S10')
        stachans[...] = ascii_stachans
        u_group = f.create_group("u")
        for i, u in enumerate(self.u):
            uset = u_group.create_dataset(name="u_" + str(i),
                                          shape=u.shape, dtype=u.dtype)
            uset[...] = u
        u_group.attrs['length'] = len(self.u)
        sigma_group = f.create_group("sigma")
        for i, sigma in enumerate(self.sigma):
            sigmaset = sigma_group.create_dataset(name="sigma_" + str(i),
                                                  shape=sigma.shape,
                                                  dtype=sigma.dtype)
            sigmaset[...] = sigma
        sigma_group.attrs['length'] = len(self.sigma)
        v_group = f.create_group("v")
        for i, v in enumerate(self.v):
            vset = v_group.create_dataset(name="v_" + str(i),
                                          shape=v.shape, dtype=v.dtype)
            vset[...] = v
        v_group.attrs['length'] = len(self.v)
        f.flush()
        f.close()
        return self

    def read(self, filename):
        """
        Read detector from a file, must be HDF5 format.

        Reads a Detector object from an HDF5 file, usually created by \
        eqcorrscan.

        :type filename: str
        :param filename: Filename to save the detector to.
        """
        f = h5py.File(filename, "r")
        self.data = []
        for i in range(f['data'].attrs['length']):
            self.data.append(f['data']['data_' + str(i)].value)
        self.u = []
        for i in range(f['u'].attrs['length']):
            self.u.append(f['u']['u_' + str(i)].value)
        self.sigma = []
        for i in range(f['sigma'].attrs['length']):
            self.sigma.append(f['sigma']['sigma_' + str(i)].value)
        self.v = []
        for i in range(f['v'].attrs['length']):
            self.v.append(f['v']['v_' + str(i)].value)
        self.stachans = [tuple(stachan.decode('ascii').split('.'))
                         for stachan in f['stachans'].value]
        self.dimension = f['data'].attrs['dimension']
        self.filt_order = f['data'].attrs['filt_order']
        self.highcut = f['data'].attrs['highcut']
        self.lowcut = f['data'].attrs['lowcut']
        self.multiplex = bool(f['data'].attrs['multiplex'])
        self.sampling_rate = f['data'].attrs['sampling_rate']
        if isinstance(f['data'].attrs['name'], str):
            self.name = f['data'].attrs['name']
        else:
            self.name = f['data'].attrs['name'].decode('ascii')
        return self

    def plot(self, stachans='all', size=(10, 7), show=True):
        """
        Plot the output basis vectors for the detector at the given dimension.

        Corresponds to the first n horizontal vectors of the V matrix.

        :type stachans: list
        :param stachans: list of tuples of station, channel pairs to plot.
        :type stachans: list
        :param stachans: List of tuples of (station, channel) to use.  Can set\
            to 'all' to use all the station-channel pairs available. If \
            detector is multiplexed, will just plot that.
        :type size: tuple
        :param size: Figure size.
        :type show: bool
        :param show: Whether or not to show the figure.

        :returns: Figure
        :rtype: matplotlib.pyplot.Figure

        .. Note::
            See :func:`eqcorrscan.utils.plotting.subspace_detector_plot`
            for example.
        """
        return subspace_detector_plot(detector=self, stachans=stachans,
                                      size=size, show=show)


def _detect(detector, st, threshold, trig_int, moveout=0, min_trig=0,
            process=True, extract_detections=False, cores=1):
    """
    Detect within continuous data using the subspace method.

    Not to be called directly, use the detector.detect method.

    :type detector: eqcorrscan.core.subspace.Detector
    :param detector: Detector to use.
    :type st: obspy.core.stream.Stream
    :param st: Un-processed stream to detect within using the subspace \
        detector
    :type threshold: float
    :param threshold: Threshold value for detections between 0-1
    :type trig_int: float
    :param trig_int: Minimum trigger interval in seconds.
    :type moveout: float
    :param moveout: Maximum allowable moveout window for non-multiplexed,
        network detection.  See note.
    :type min_trig: int
    :param min_trig: Minimum number of stations exceeding threshold for \
        non-multiplexed, network detection. See note.
    :type process: bool
    :param process: Whether or not to process the stream according to the \
        parameters defined by the detector.  Default is to process the \
        data (True).
    :type extract_detections: bool
    :param extract_detections: Whether to extract waveforms for each \
        detection or not, if true will return detections and streams.

    :return: list of detections
    :rtype: list of eqcorrscan.core.match_filter.Detection
    """
    detections = []
    # First process the stream
    if process:
        Logger.info('Processing Stream')
        stream, stachans = _subspace_process(
            streams=[st.copy()], lowcut=detector.lowcut,
            highcut=detector.highcut, filt_order=detector.filt_order,
            sampling_rate=detector.sampling_rate, multiplex=detector.multiplex,
            stachans=detector.stachans, parallel=True, align=False,
            shift_len=None, reject=False, cores=cores)
    else:
        # Check the sampling rate at the very least
        for tr in st:
            if not tr.stats.sampling_rate == detector.sampling_rate:
                raise ValueError('Sampling rates do not match.')
        stream = [st]
        stachans = detector.stachans
    outtic = time.clock()
    # If multiplexed, how many samples do we increment by?
    if detector.multiplex:
        Nc = len(detector.stachans)
    else:
        Nc = 1
    # Here do all ffts
    fft_vars = _do_ffts(detector, stream, Nc)
    Logger.info('Computing detection statistics')
    Logger.info('Preallocating stats matrix')
    stats = np.zeros((len(stream[0]),
                      (len(stream[0][0]) // Nc) - (fft_vars[4] // Nc) + 1))
    for det_freq, data_freq_sq, data_freq, i in zip(fft_vars[0], fft_vars[1],
                                                    fft_vars[2],
                                                    np.arange(len(stream[0]))):
        # Calculate det_statistic in frequency domain
        stats[i] = _det_stat_freq(det_freq, data_freq_sq, data_freq,
                                  fft_vars[3], Nc, fft_vars[4], fft_vars[5])
        Logger.info('Stats matrix is shape %s' % str(stats[i].shape))
    trig_int_samples = detector.sampling_rate * trig_int
    Logger.info('Finding peaks')
    peaks = []
    for i in range(len(stream[0])):
        peaks.append(findpeaks.find_peaks2_short(
            arr=stats[i], thresh=threshold, trig_int=trig_int_samples))
    if not detector.multiplex:
        # Conduct network coincidence triggering
        peaks = findpeaks.coin_trig(
            peaks=peaks, samp_rate=detector.sampling_rate, moveout=moveout,
            min_trig=min_trig, stachans=stachans, trig_int=trig_int)
    else:
        peaks = peaks[0]
    if len(peaks) > 0:
        for peak in peaks:
            detecttime = st[0].stats.starttime + \
                (peak[1] / detector.sampling_rate)
            rid = ResourceIdentifier(
                id=detector.name + '_' + str(detecttime), prefix='smi:local')
            ev = Event(resource_id=rid)
            cr_i = CreationInfo(
                author='EQcorrscan', creation_time=UTCDateTime())
            ev.creation_info = cr_i
            # All detection info in Comments for lack of a better idea
            thresh_str = 'threshold=' + str(threshold)
            ccc_str = 'detect_val=' + str(peak[0])
            used_chans = 'channels used: ' +\
                ' '.join([str(pair) for pair in detector.stachans])
            ev.comments.append(Comment(text=thresh_str))
            ev.comments.append(Comment(text=ccc_str))
            ev.comments.append(Comment(text=used_chans))
            for stachan in detector.stachans:
                tr = st.select(station=stachan[0], channel=stachan[1])
                if tr:
                    net_code = tr[0].stats.network
                else:
                    net_code = ''
                pick_tm = detecttime
                wv_id = WaveformStreamID(
                    network_code=net_code, station_code=stachan[0],
                    channel_code=stachan[1])
                ev.picks.append(Pick(time=pick_tm, waveform_id=wv_id))
            detections.append(
                Detection(template_name=detector.name, detect_time=detecttime,
                          no_chans=len(detector.stachans), detect_val=peak[0],
                          threshold=threshold, typeofdet='subspace',
                          threshold_type='abs', threshold_input=threshold,
                          chans=detector.stachans, event=ev))
    outtoc = time.clock()
    Logger.info('Detection took %s seconds' % str(outtoc - outtic))
    if extract_detections:
        detection_streams = extract_from_stream(st, detections)
        return detections, detection_streams
    return detections


def _do_ffts(detector, stream, Nc):
    """
    Perform ffts on data, detector and denominator boxcar

    :type detector: eqcorrscan.core.subspace.Detector
    :param detector: Detector object for doing detecting
    :type stream: list of obspy.core.stream.Stream
    :param stream: List of streams processed according to detector
    :type Nc: int
    :param Nc: Number of channels in data. 1 for non-multiplexed

    :return: list of time-reversed detector(s) in freq domain
    :rtype: list
    :return: list of squared data stream(s) in freq domain
    :rtype: list
    :return: list of data stream(s) in freq domain
    :return: detector-length boxcar in freq domain
    :rtype: numpy.ndarray
    :return: length of detector
    :rtype: int
    :return: length of data
    :rtype: int
    """
    min_fftlen = int(stream[0][0].data.shape[0] +
                     detector.data[0].shape[0] - Nc)
    fftlen = scipy.fftpack.next_fast_len(min_fftlen)
    mplen = stream[0][0].data.shape[0]
    ulen = detector.data[0].shape[0]
    num_st_fd = [np.fft.rfft(tr.data, n=fftlen)
                 for tr in stream[0]]
    denom_st_fd = [np.fft.rfft(np.square(tr.data), n=fftlen)
                   for tr in stream[0]]
    # Frequency domain of boxcar
    w = np.fft.rfft(np.ones(detector.data[0].shape[0]),
                    n=fftlen)
    # This should go into the detector object as in Detex
    detector_fd = []
    for dat_mat in detector.data:
        detector_fd.append(np.array([np.fft.rfft(col[::-1], n=fftlen)
                                     for col in dat_mat.T]))
    return detector_fd, denom_st_fd, num_st_fd, w, ulen, mplen


def _det_stat_freq(det_freq, data_freq_sq, data_freq, w, Nc, ulen, mplen):
    """
    Compute detection statistic in the frequency domain

    :type det_freq: numpy.ndarray
    :param det_freq: detector in freq domain
    :type data_freq_sq: numpy.ndarray
    :param data_freq_sq: squared data in freq domain
    :type data_freq: numpy.ndarray
    :param data_freq: data in freq domain
    :type w: numpy.ndarray
    :param w: boxcar in freq domain
    :type Nc: int
    :param Nc: number of channels in data stream
    :type ulen: int
    :param ulen: length of detector
    :type mplen: int
    :param mplen: length of data

    :return: Array of detection statistics
    :rtype: numpy.ndarray
    """
    num_cor = np.multiply(det_freq, data_freq)  # Numerator convolution
    den_cor = np.multiply(w, data_freq_sq)  # Denominator convolution
    # Do inverse fft
    # First and last Nt - 1 samples are invalid; clip them off
    num_ifft = np.real(np.fft.irfft(num_cor))[:, ulen-1:mplen:Nc]
    denominator = np.real(np.fft.irfft(den_cor))[ulen-1:mplen:Nc]
    # Ratio of projected to envelope energy = det_stat across all channels
    result = np.sum(np.square(num_ifft), axis=0) / denominator
    return result


def _subspace_process(streams, lowcut, highcut, filt_order, sampling_rate,
                      multiplex, align, shift_len, reject, no_missed=True,
                      stachans=None, parallel=False, plot=False, cores=1):
    """
    Process stream data, internal function.

    :type streams: list
    :param streams: List of obspy.core.stream.Stream to be used to \
        generate the subspace detector.  These should be pre-clustered \
        and aligned.
    :type lowcut: float
    :param lowcut: Lowcut in Hz, can be None to not apply filter
    :type highcut: float
    :param highcut: Highcut in Hz, can be None to not apply filter
    :type filt_order: int
    :param filt_order: Number of corners for filter.
    :type sampling_rate: float
    :param sampling_rate: Desired sampling rate in Hz
    :type multiplex: bool
    :param multiplex: Whether to multiplex the data or not.  Data are \
        multiplexed according to the method of Harris, see the multi \
        function for details.
    :type stachans: list of tuple
    :param stachans: list of tuples of (station, channel) to use.
    :type align: bool
    :param align: Whether to align the data or not - needs to be done \
        at some point
    :type shift_len: float
    :param shift_len: Maximum shift allowed for alignment in seconds.
    :type reject: float
    :param reject: Minimum correlation for traces, only used if align=True.
    :type no_missed: bool
    :param: no_missed: Reject streams with missed traces, defaults to True. \
        A missing trace from lots of events will reduce the quality of the \
        subspace detector if multiplexed.  Only used when multi is set to True.
    :type plot: bool
    :param plot: Passed down to align traces - used to check alignment process.

    :return: Processed streams
    :rtype: list
    :return: Station, channel pairs in order
    :rtype: list of tuple
    :return: List of delays
    :rtype: list
    """
    from multiprocessing import Pool, cpu_count
    processed_streams = []
    if not stachans:
        input_stachans = list(set([(tr.stats.station, tr.stats.channel)
                                   for st in streams for tr in st.sort()]))
    else:
        input_stachans = stachans
    input_stachans.sort()  # Make sure stations and channels are in order
    # Check that all channels are the same length in seconds
    first_length = len(streams[0][0].data) /\
        streams[0][0].stats.sampling_rate
    for st in streams:
        for tr in st:
            if not len(tr) / tr.stats.sampling_rate == first_length:
                raise IOError(
                    'All channels of all streams must be the same length')
    for st in streams:
        if not parallel:
            processed_stream = Stream()
            for stachan in input_stachans:
                dummy, tr = _internal_process(
                    st=st, lowcut=lowcut, highcut=highcut,
                    filt_order=filt_order, sampling_rate=sampling_rate,
                    first_length=first_length, stachan=stachan)
                processed_stream += tr
            processed_streams.append(processed_stream)
        else:
            pool = Pool(processes=min(cores, cpu_count()))
            results = [pool.apply_async(
                _internal_process, (st,),
                {'lowcut': lowcut, 'highcut': highcut,
                 'filt_order': filt_order, 'sampling_rate': sampling_rate,
                 'first_length': first_length, 'stachan': stachan,
                 'i': i}) for i, stachan in enumerate(input_stachans)]
            pool.close()
            try:
                processed_stream = [p.get() for p in results]
            except KeyboardInterrupt as e:  # pragma: no cover
                pool.terminate()
                raise e
            pool.join()
            processed_stream.sort(key=lambda tup: tup[0])
            processed_stream = Stream([p[1] for p in processed_stream])
            processed_streams.append(processed_stream)
        if no_missed and multiplex:
            for tr in processed_stream:
                if np.count_nonzero(tr.data) == 0:
                    processed_streams.remove(processed_stream)
                    Logger.info('Removed stream with empty trace')
                    break
    if align:
        processed_streams = align_design(
            design_set=processed_streams, shift_len=shift_len, reject=reject,
            multiplex=multiplex, plot=plot, no_missed=no_missed)
    output_streams = []
    for processed_stream in processed_streams:
        if len(processed_stream) == 0:
            # If we have removed all of the traces then onwards!
            continue
        # Need to order the stream according to input_stachans
        _st = Stream()
        for stachan in input_stachans:
            tr = processed_stream.select(
                station=stachan[0], channel=stachan[1])
            if len(tr) >= 1:
                _st += tr[0]
            elif multiplex and len(tr) == 0:
                raise IndexError(
                    'Missing data for %s.%s' % (stachan[0], stachan[1]))
        if multiplex:
            st = multi(stream=_st)
            st = Stream(Trace(st))
            st[0].stats.station = 'Multi'
            st[0].stats.sampling_rate = sampling_rate
        else:
            st = _st
        for tr in st:
            # Normalize the data
            norm = np.linalg.norm(tr.data)
            if not norm == 0:
                tr.data /= norm
        output_streams.append(st)
    return output_streams, input_stachans


def _internal_process(st, lowcut, highcut, filt_order, sampling_rate,
                      first_length, stachan, i=0):
    tr = st.select(station=stachan[0], channel=stachan[1])
    if len(tr) == 0:
        tr = Trace(np.zeros(int(first_length * sampling_rate)))
        tr.stats.station = stachan[0]
        tr.stats.channel = stachan[1]
        tr.stats.sampling_rate = sampling_rate
        tr.stats.starttime = st[0].stats.starttime  # Do this to make more
        # sensible plots
        Logger.warning('Padding stream with zero trace for ' +
                       'station ' + stachan[0] + '.' + stachan[1])
    elif len(tr) == 1:
        tr = tr[0]
        tr.detrend('simple')
        tr = pre_processing.process(
            tr=tr, lowcut=lowcut, highcut=highcut, filt_order=filt_order,
            samp_rate=sampling_rate, seisan_chan_names=False)
    else:
        raise IOError('Multiple channels for {0}.{1} in a single design '
                      'stream.'.format(stachan[0], stachan[1]))
    return i, tr


def read_detector(filename):
    """
    Read detector from a filename.

    :type filename: str
    :param filename: Filename to save the detector to.

    :return: Detector object
    :rtype: eqcorrscan.core.subspace.Detector
    """
    detector = Detector()
    detector.read(filename=filename)
    return detector


def multi(stream):
    """
    Internal multiplexer for multiplex_detect.

    :type stream: obspy.core.stream.Stream
    :param stream: Stream to multiplex

    :return: trace of multiplexed data
    :rtype: obspy.core.trace.Trace

    .. Note: Requires all channels to be the same length.

    Maps a standard multiplexed stream of seismic data to a single traces of \
    multiplexed data as follows:

    Input:
    x = [x1, x2, x3, ...]
    y = [y1, y2, y3, ...]
    z = [z1, z2, z3, ...]

    Output:
    xyz = [x1, y1, z1, x2, y2, z2, x3, y3, z3, ...]
    """
    stack = stream[0].data
    for tr in stream[1:]:
        stack = np.dstack(np.array([stack, tr.data]))
    multiplex = stack.reshape(stack.size, )
    return multiplex


def align_design(design_set, shift_len, reject, multiplex, no_missed=True,
                 plot=False):
    """
    Align individual traces within streams of the design set.

    Perform before Detector.construct to align traces before computing the \
    singular value decomposition.

    :type design_set: list
    :param design_set: List of obspy.core.stream.Stream to be aligned
    :type shift_len: float
    :param shift_len: Maximum shift (plus/minus) in seconds.
    :type reject: float
    :param reject: Minimum correlation for traces, only used if align=True.
    :type multiplex: bool
    :param multiplex: If you are going to multiplex the data, then there has \
        to be data for all channels, so we will pad with zeros, otherwise \
        there is no need.
    :type no_missed: bool
    :param: no_missed: Reject streams with missed traces, defaults to True. \
        A missing trace from lots of events will reduce the quality of the \
        subspace detector if multiplexed.  Only used when multi is set to True.
    :type plot: bool
    :param plot: Whether to plot the aligned traces as we go or not.

    :rtype: list
    :return: List of obspy.core.stream.Stream of aligned streams

    .. Note:: Assumes only one trace for each channel for each stream in the \
        design_set. If more are present will only use the first one.

    .. Note:: Will cut all traces to be the same length as required for the \
        svd, this length will be the shortest trace length - 2 * shift_len
    """
    trace_lengths = [tr.stats.endtime - tr.stats.starttime for st in design_set
                     for tr in st]
    clip_len = min(trace_lengths) - (2 * shift_len)
    stachans = list(set([(tr.stats.station, tr.stats.channel)
                         for st in design_set for tr in st]))
    remove_set = []
    for stachan in stachans:
        trace_list = []
        trace_ids = []
        for i, st in enumerate(design_set):
            tr = st.select(station=stachan[0], channel=stachan[1])
            if len(tr) > 0:
                trace_list.append(tr[0])
                trace_ids.append(i)
            if len(tr) > 1:
                Logger.warning(
                    'Too many matches for %s %s' % (stachan[0], stachan[1]))
        shift_len_samples = int(shift_len * trace_list[0].stats.sampling_rate)
        shifts, cccs = stacking.align_traces(
            trace_list=trace_list, shift_len=shift_len_samples, positive=True)
        for i, shift in enumerate(shifts):
            st = design_set[trace_ids[i]]
            start_t = st.select(
                station=stachan[0], channel=stachan[1])[0].stats.starttime
            start_t += shift_len
            start_t -= shift
            st.select(
                station=stachan[0], channel=stachan[1])[0].trim(
                start_t, start_t + clip_len)
            if cccs[i] < reject:
                if multiplex and not no_missed:
                    st.select(station=stachan[0],
                              channel=stachan[1])[0].data = np.zeros(
                        int(clip_len * (st.select(
                            station=stachan[0],
                            channel=stachan[1])[0].stats.sampling_rate) + 1))
                    Logger.warning('Padding stream with zero trace for ' +
                                   'station ' + stachan[0] + '.' + stachan[1])
                    Logger.debug('zero padding')
                elif multiplex and no_missed:
                    remove_set.append(st)
                    Logger.warning('Will remove stream due to low-correlation')
                    continue
                else:
                    st.remove(st.select(station=stachan[0],
                              channel=stachan[1])[0])
                    Logger.warning(
                        'Removed channel with correlation at %s' % cccs[i])
                    continue
    if no_missed:
        for st in remove_set:
            if st in design_set:
                design_set.remove(st)
    if plot:
        for stachan in stachans:
            trace_list = []
            for st in design_set:
                tr = st.select(station=stachan[0], channel=stachan[1])
                if len(tr) > 0:
                    trace_list.append(tr[0])
            if len(trace_list) > 1:
                plotting.multi_trace_plot(traces=trace_list, corr=True,
                                          stack=None, title='.'.join(stachan))
            else:
                Logger.error(
                    'No plot for you, only one trace left after rejection')
    return design_set


def subspace_detect(detectors, stream, threshold, trig_int, moveout=0,
                    min_trig=1, parallel=True, num_cores=None):
    """
    Conduct subspace detection with chosen detectors.

    :type detectors: list
    :param detectors:
        list of :class:`eqcorrscan.core.subspace.Detector` to be used
        for detection.
    :type stream: obspy.core.stream.Stream
    :param stream: Stream to detect within.
    :type threshold: float
    :param threshold:
        Threshold between 0 and 1 for detection, see :func:`Detector.detect`
    :type trig_int: float
    :param trig_int: Minimum trigger interval in seconds.
    :type moveout: float
    :param moveout:
        Maximum allowable moveout window for non-multiplexed, network
        detection.  See note.
    :type min_trig: int
    :param min_trig:
        Minimum number of stations exceeding threshold for non-multiplexed,
        network detection. See note in :func:`Detector.detect`.
    :type parallel: bool
    :param parallel: Whether to run detectors in parallel in groups.
    :type num_cores: int
    :param num_cores:
        How many cpu cores to use if parallel==True. If set to None (default),
        will use all available cores.

    :rtype: list
    :return:
        List of :class:`eqcorrscan.core.match_filter.Detection` detections.

    .. Note::
        This will loop through your detectors using their detect method.
        If the detectors are multiplexed it will run groups of detectors with
        the same channels at the same time.
    """
    from multiprocessing import Pool, cpu_count
    # First check that detector parameters are the same
    parameters = []
    detections = []
    for detector in detectors:
        parameter = (detector.lowcut, detector.highcut,
                     detector.filt_order, detector.sampling_rate,
                     detector.multiplex, detector.stachans)
        if parameter not in parameters:
            parameters.append(parameter)
    for parameter_set in parameters:
        parameter_detectors = []
        for detector in detectors:
            det_par = (detector.lowcut, detector.highcut, detector.filt_order,
                       detector.sampling_rate, detector.multiplex,
                       detector.stachans)
            if det_par == parameter_set:
                parameter_detectors.append(detector)
        stream, stachans = \
            _subspace_process(
                streams=[stream.copy()], lowcut=parameter_set[0],
                highcut=parameter_set[1], filt_order=parameter_set[2],
                sampling_rate=parameter_set[3], multiplex=parameter_set[4],
                stachans=parameter_set[5], parallel=True, align=False,
                shift_len=None, reject=False)
        if not parallel:
            for detector in parameter_detectors:
                detections += _detect(
                    detector=detector, st=stream[0], threshold=threshold,
                    trig_int=trig_int, moveout=moveout, min_trig=min_trig,
                    process=False, extract_detections=False)
        else:
            if num_cores:
                ncores = num_cores
            else:
                ncores = cpu_count()
            pool = Pool(processes=ncores)
            results = [pool.apply_async(
                _detect, args=(detector, stream[0], threshold, trig_int,
                               moveout, min_trig, False, False))
                       for detector in parameter_detectors]
            pool.close()
            try:
                _detections = [p.get() for p in results]
            except KeyboardInterrupt as e:  # pragma: no cover
                pool.terminate()
                raise e
            pool.join()
            for d in _detections:
                if isinstance(d, list):
                    detections += d
                else:
                    detections.append(d)
    return detections



"""
Functions for testing the utils.pre_processing functions.
"""
import unittest
import os
import numpy as np
import glob

from random import shuffle
from copy import deepcopy
from obspy import read, Trace, UTCDateTime, Stream

from eqcorrscan.utils.pre_processing import (
    process, dayproc, shortproc, _check_daylong, _prep_data_for_correlation)


class TestPreProcessing(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        testing_path = os.path.join(
            os.path.abspath(os.path.dirname(__file__)), 'test_data',
            'day_vols', 'Y2012', 'R086.01', '*')
        cls.st = read(testing_path)
        cls.day_start = cls.st[0].stats.starttime.date
        cls.short_stream = cls.st.copy().trim(cls.st[0].stats.starttime,
                                              cls.st[0].stats.starttime + 3600)
        cls.instart = cls.short_stream[0].stats.starttime
        cls.inend = cls.short_stream[0].stats.endtime
        cls.nchans = len(cls.short_stream)
        cls.gap_starttime = cls.st[0].stats.starttime + 1800
        cls.gap_endtime = cls.st[0].stats.starttime + 1900
        tr1 = cls.st[0].copy().trim(
            cls.st[0].stats.starttime, cls.gap_starttime)
        tr2 = cls.st[0].copy().trim(
            cls.gap_endtime, cls.st[0].stats.starttime + 3600)
        cls.gappy_trace = tr1 + tr2

    def test_daylong_checks(self):
        """Test that the data are day-long."""
        self.assertTrue(_check_daylong(self.st[0]))
        not_daylong = self.st[0].copy().trim(self.st[0].stats.starttime,
                                             self.st[0].stats.starttime + 3600)
        not_daylong.data = np.append(
            not_daylong.data, np.zeros(
                3602 * int(self.st[0].stats.sampling_rate)))
        self.assertFalse(_check_daylong(not_daylong))

    def test_shortproc(self):
        """Test the short-proc processing method."""
        processed = shortproc(
            self.short_stream.copy(), lowcut=0.1, highcut=0.4, filt_order=4,
            samp_rate=1, parallel=False, num_cores=False,
            starttime=None, endtime=None)
        self.assertEqual(len(processed), self.nchans)
        for tr in processed:
            self.assertEqual(self.instart, tr.stats.starttime)
            self.assertEqual(self.inend, tr.stats.endtime)

    def test_filter_error(self):
        """Check that we don't allow filtering above the nyquist."""
        with self.assertRaises(IOError):
            shortproc(self.short_stream.copy(), lowcut=0.1, highcut=0.6,
                      filt_order=4, samp_rate=1, parallel=False,
                      num_cores=False, starttime=None, endtime=None)

    def test_shortproc_set_start(self):
        """Check that shortproc trims properly."""
        processed = shortproc(
            self.short_stream.copy(), lowcut=0.1, highcut=0.4, filt_order=4,
            samp_rate=1, parallel=False, num_cores=False,
            starttime=self.short_stream[0].stats.starttime + 2, endtime=None)
        self.assertEqual(len(processed), self.nchans)
        for tr in processed:
            self.assertEqual(self.instart + 2, tr.stats.starttime)
            self.assertEqual(self.inend, tr.stats.endtime)

    def test_shortproc_set_end(self):
        """Check that shortproc trims properly."""
        processed = shortproc(
            self.short_stream.copy(), lowcut=0.1, highcut=0.4, filt_order=4,
            samp_rate=1, parallel=False, num_cores=False,
            starttime=None, endtime=self.short_stream[0].stats.endtime - 2)
        self.assertEqual(len(processed), self.nchans)
        for tr in processed:
            self.assertEqual(self.instart, tr.stats.starttime)
            self.assertEqual(self.inend - 2, tr.stats.endtime)

    def test_shortproc_set_start_and_end(self):
        """Check that shortproc trims properly."""
        processed = shortproc(
            self.short_stream.copy(), lowcut=0.1, highcut=0.4, filt_order=4,
            samp_rate=1, parallel=False, num_cores=False,
            starttime=self.short_stream[0].stats.starttime + 2,
            endtime=self.short_stream[0].stats.endtime - 2)
        self.assertEqual(len(processed), self.nchans)
        for tr in processed:
            self.assertEqual(self.instart + 2, tr.stats.starttime)
            self.assertEqual(self.inend - 3, tr.stats.endtime)

    def test_trace_as_argument(self):
        """
        Check that we can cope with a trace, and that a trace is returned.
        """
        processed = shortproc(
            self.short_stream.copy()[0], lowcut=0.1, highcut=0.4, filt_order=4,
            samp_rate=1, parallel=False, num_cores=False,
            starttime=None, endtime=None)
        self.assertTrue(isinstance(processed, Trace))
        self.assertEqual(self.instart, processed.stats.starttime)
        self.assertEqual(self.inend, processed.stats.endtime)

    def test_parallel(self):
        """Test the parallel implementation."""
        processed = shortproc(
            self.short_stream.copy(), lowcut=0.1, highcut=0.4, filt_order=4,
            samp_rate=1, parallel=True, num_cores=2,
            starttime=None, endtime=None)
        self.assertEqual(len(processed), self.nchans)
        for tr in processed:
            self.assertEqual(self.instart, tr.stats.starttime)
            self.assertEqual(self.inend, tr.stats.endtime)

    def test_parallel_core_unset(self):
        """Test the parallel implementation without num_cores set."""
        processed = shortproc(
            self.short_stream.copy(), lowcut=0.1, highcut=0.4, filt_order=4,
            samp_rate=1, parallel=True, num_cores=False,
            starttime=None, endtime=None)
        self.assertEqual(len(processed), self.nchans)
        for tr in processed:
            self.assertEqual(self.instart, tr.stats.starttime)
            self.assertEqual(self.inend, tr.stats.endtime)

    def test_dayproc(self):
        """Test a straight-forward day processing implementation."""
        processed = dayproc(
            st=self.st.copy(), lowcut=0.1, highcut=0.4, filt_order=3,
            samp_rate=1, starttime=self.day_start, parallel=True,
            num_cores=2)
        self.assertEqual(len(processed), self.nchans)
        for tr in processed:
            self.assertEqual(UTCDateTime(self.day_start), tr.stats.starttime)
            self.assertEqual(tr.stats.npts, 86400)

    def test_dayproc_trace(self):
        """
        Test a straight-forward day processing implementation with a Trace.
        """
        processed = dayproc(
            st=self.st[0].copy(), lowcut=0.1, highcut=0.4, filt_order=3,
            samp_rate=1, starttime=self.day_start, parallel=True,
            num_cores=2)
        self.assertTrue(isinstance(processed, Trace))
        self.assertEqual(UTCDateTime(self.day_start),
                         processed.stats.starttime)
        self.assertEqual(processed.stats.npts, 86400)

    def test_dayproc_nyquist_error(self):
        """Test a failing day processing."""
        with self.assertRaises(IOError):
            dayproc(st=self.st.copy(), lowcut=0.1, highcut=0.6, filt_order=3,
                    samp_rate=1, starttime=self.day_start,
                    parallel=True, num_cores=2)

    def test_dayproc_serial(self):
        """Test the serial implementation of dayproc."""
        processed = dayproc(st=self.st.copy(), lowcut=0.1, highcut=0.4,
                            filt_order=3, samp_rate=1,
                            starttime=self.day_start, parallel=False,
                            num_cores=2)
        self.assertEqual(len(processed), self.nchans)
        for tr in processed:
            self.assertEqual(UTCDateTime(self.day_start), tr.stats.starttime)
            self.assertEqual(tr.stats.npts, 86400)

    def test_dayproc_parallel_cores_unset(self):
        """Test a straight-forward day processing implementation."""
        processed = dayproc(
            st=self.st.copy(), lowcut=0.1, highcut=0.4, filt_order=3,
            samp_rate=1, starttime=self.day_start, parallel=True,
            num_cores=False)
        self.assertEqual(len(processed), self.nchans)
        for tr in processed:
            self.assertEqual(UTCDateTime(self.day_start), tr.stats.starttime)
            self.assertEqual(tr.stats.npts, 86400)

    def test_process(self):
        """Test a basic process implementation."""
        processed = process(tr=self.st[0].copy(), lowcut=0.1, highcut=0.4,
                            filt_order=3, samp_rate=1,
                            starttime=False, clip=False, length=86400,
                            seisan_chan_names=True, ignore_length=False)
        self.assertEqual(processed.stats.npts, 86400)

    def test_process_datetime(self):
        """Test a basic process implementation."""
        processed = process(tr=self.st[0].copy(), lowcut=0.1, highcut=0.4,
                            filt_order=3, samp_rate=1,
                            starttime=self.day_start, clip=False, length=86400,
                            seisan_chan_names=True, ignore_length=False)
        self.assertEqual(processed.stats.npts, 86400)

    def test_process_nyquist_fail(self):
        """Test a nyquist error is raised."""
        with self.assertRaises(IOError):
            process(tr=self.st[0].copy(), lowcut=0.1, highcut=0.6,
                    filt_order=3, samp_rate=1, starttime=False, clip=False,
                    length=86400, seisan_chan_names=True, ignore_length=False)

    def test_process_bad_data(self):
        """Check that we won't allow data that are mostly zeros."""
        not_daylong = self.st[0].copy().trim(self.st[0].stats.starttime,
                                             self.st[0].stats.starttime + 3600)
        not_daylong.data = np.append(
            not_daylong.data, np.zeros(
                3602 * int(self.st[0].stats.sampling_rate)))
        with self.assertRaises(ValueError):
            process(tr=not_daylong, lowcut=0.1, highcut=0.4,
                    filt_order=3, samp_rate=1,
                    starttime=False, clip=False, length=86400,
                    seisan_chan_names=True, ignore_length=False)

    def test_short_data_fail(self):
        """Check that we don't allow too much missing data."""
        with self.assertRaises(NotImplementedError):
            process(tr=self.st[0].copy().
                    trim(endtime=self.st[0].stats.endtime - 18000), lowcut=0.1,
                    highcut=0.4, filt_order=3, samp_rate=1,
                    starttime=self.day_start, clip=True, length=86400,
                    seisan_chan_names=True, ignore_length=False)

    def test_short_data_pass(self):
        """Check that we do allow missing data if ignore_length is True."""
        processed = process(
            tr=self.st[0].copy().trim(endtime=self.
                                      st[0].stats.endtime - 18000), lowcut=0.1,
            highcut=0.4, filt_order=3, samp_rate=1,
            starttime=self.day_start, clip=True, length=86400,
            seisan_chan_names=True, ignore_length=True)
        self.assertEqual(processed.stats.npts, 86400)

    def test_highcut_debug(self):
        """Test a basic process implementation with just a highcut"""
        processed = process(tr=self.st[0].copy(), lowcut=None, highcut=0.4,
                            filt_order=3, samp_rate=1,
                            starttime=False, clip=False, length=86400,
                            seisan_chan_names=True, ignore_length=False)
        self.assertEqual(processed.stats.npts, 86400)

    def test_lowcut_debug(self):
        """Test a basic process implementation with just a highcut"""
        processed = process(tr=self.st[0].copy(), lowcut=0.1, highcut=None,
                            filt_order=3, samp_rate=1,
                            starttime=False, clip=False, length=86400,
                            seisan_chan_names=True, ignore_length=False)
        self.assertEqual(processed.stats.npts, 86400)

    def test_masked_trace(self):
        """Test that processing a masked array works."""
        tr = self.gappy_trace
        processed = process(tr=tr, lowcut=0.1, highcut=0.4,
                            filt_order=3, samp_rate=1,
                            starttime=False, clip=False, length=3600,
                            seisan_chan_names=True, ignore_length=False)
        self.assertEqual(processed.stats.npts, 3601)
        self.assertFalse(isinstance(processed.data, np.ma.MaskedArray))
        self.assertTrue(np.all(
            processed.trim(self.gap_starttime, self.gap_endtime).data) == 0)

    def test_masked_trace_no_fill(self):
        """Test that processing a masked array without filling gaps works."""
        tr = self.gappy_trace
        processed = process(tr=tr, lowcut=0.1, highcut=0.4,
                            filt_order=3, samp_rate=1,
                            starttime=False, clip=False, length=3600,
                            seisan_chan_names=True, ignore_length=False,
                            fill_gaps=False)
        self.assertEqual(processed.stats.npts, 3601)
        self.assertTrue(isinstance(processed.data, np.ma.MaskedArray))

    def test_masked_array_resample(self):
        """Test that processing and resampling a masked array works."""
        tr = self.gappy_trace
        processed = process(tr=tr, lowcut=0.1, highcut=0.2,
                            filt_order=3, samp_rate=0.5,
                            starttime=False, clip=False, length=3600,
                            seisan_chan_names=True, ignore_length=False)
        self.assertEqual(processed.stats.npts, 1800)
        self.assertTrue(np.all(
            processed.trim(self.gap_starttime, self.gap_endtime).data) == 0)

    def test_gap_overlength(self):
        """ Test that gappy data that are too long are processed correctly."""
        tr_after = self.gappy_trace.copy()
        tr_after.stats.starttime += 3600
        tr_before = self.gappy_trace.copy()
        tr_before.stats.starttime -= 3600
        tr = tr_before + self.gappy_trace + tr_after
        processed = process(
            tr=tr, lowcut=0.1, highcut=0.4, filt_order=3, samp_rate=1,
            starttime=self.gappy_trace.stats.starttime, clip=True, length=3600,
            seisan_chan_names=True, ignore_length=False)
        self.assertEqual(processed.stats.npts, 3600)
        self.assertFalse(isinstance(processed.data, np.ma.MaskedArray))
        self.assertTrue(np.all(
            processed.trim(self.gap_starttime, self.gap_endtime).data) == 0)


class TestDataPrep(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        testing_path = os.path.join(
            os.path.abspath(os.path.dirname(__file__)), 'test_data',
            'similar_events_processed')
        stream_files = glob.glob(os.path.join(testing_path, '*'))
        stream_list = [read(stream_file)
                       for stream_file in stream_files]
        cls.stream_list = stream_list

    def test_order_irrevelant(self):
        """ Early on the order of streams input was affecting output. """
        st1 = self.stream_list[0].copy()

        ordered_streams = sorted(
            self.stream_list, key=lambda s: s[0].stats.starttime)
        shuffled_streams = deepcopy(ordered_streams)
        shuffle(shuffled_streams)

        order_prepped_st1, order_prepped_streams = _prep_data_for_correlation(
            stream=st1, templates=ordered_streams, force_stream_epoch=False)
        shuff_prepped_st1, shuff_prepped_streams = _prep_data_for_correlation(
            stream=st1, templates=shuffled_streams, force_stream_epoch=False)

        self.assertEqual(order_prepped_st1, shuff_prepped_st1)
        shuff_prepped_streams = sorted(
            shuff_prepped_streams, key=lambda s: s[0].stats.starttime)

        for ordered, shuffled in zip(order_prepped_streams,
                                     shuff_prepped_streams):
            for tr_ordered, tr_shuffled in zip(ordered, shuffled):
                self.assertEqual(tr_ordered, tr_shuffled)

    def test_correct_padding(self):
        """check that nans are put where they are supposed to be """
        _, prepped_streams = _prep_data_for_correlation(
            stream=self.stream_list[0], templates=deepcopy(self.stream_list),
            force_stream_epoch=False)

        for original, prepped in zip(self.stream_list, prepped_streams):
            for tr in prepped:
                original_tr = original.select(id=tr.id)
                if len(original_tr) == 0:
                    self.assertTrue(np.all(np.isnan(tr.data)))
                else:
                    self.assertEqual(tr, original_tr[0])

    def test_duplicate_template_channels(self):
        """
        Check that duplicate template channels result in duplicate template
        channels but not stream channels.
        """
        templates = deepcopy(self.stream_list)
        templates.sort(key=lambda t: t[0].stats.starttime)
        stream = deepcopy(self.stream_list[0])
        additional_channel = templates[0][0]
        additional_channel.stats.starttime += 30
        templates[0] += additional_channel
        assert len(templates[0]) == 10
        continuous_data, templates = _prep_data_for_correlation(
            stream=stream, templates=templates, force_stream_epoch=True)
        for template in templates:
            assert len(template) == 10
        assert len(continuous_data) == 9

    def test_continuous_data_removal(self):
        """Check that data that should be removed are."""
        st = read()
        templates = []
        for i in range(20):
            template = Stream()
            for _tr in st[1:]:
                tr = _tr.copy()
                tr.stats.starttime += i * 100
                tr.data = np.random.randn(200)
                template += tr
            templates.append(template)
        continuous_data = st
        for tr in continuous_data:
            tr.data = np.random.randn(200 * 20)

        continuous_data, templates = _prep_data_for_correlation(
            stream=continuous_data, templates=templates,
            force_stream_epoch=True)
        self.assertEqual(len(continuous_data), 2)
        for template in templates:
            self.assertEqual(len(template), 2)

    def test_length_checking(self):
        """Check that lengths are fixed"""
        st = read()
        templates = []
        for i in range(20):
            template = Stream()
            for _tr in st:
                tr = _tr.copy()
                tr.stats.starttime += i * 100
                tr.data = np.random.randn(200)
                template += tr
            templates.append(template)
        continuous_data = st
        for i, tr in enumerate(continuous_data):
            tr.data = np.random.randn((200 * 20) + ((i + 1) * 5))
            tr.stats.starttime += (i + 1) * 5
        min_start = min([tr.stats.starttime for tr in continuous_data])
        max_end = max([tr.stats.endtime for tr in continuous_data])
        data_len = (
            (max_end - min_start) * continuous_data[0].stats.sampling_rate) + 1

        continuous_data_prepped, templates = _prep_data_for_correlation(
            stream=continuous_data.copy(), templates=templates,
            force_stream_epoch=True)

        for tr in continuous_data_prepped:
            self.assertEqual(tr.stats.npts, data_len)
            self.assertEqual(tr.stats.starttime, min_start)
            self.assertEqual(tr.stats.endtime, max_end)

        continuous_data_prepped, templates = _prep_data_for_correlation(
            stream=continuous_data, templates=templates,
            force_stream_epoch=False)

        for tr in continuous_data_prepped:
            self.assertEqual(
                tr.stats.npts, (200 * 20) + (len(continuous_data) * 5))


if __name__ == '__main__':
    unittest.main()



"""
Functions for network matched-filter detection of seismic data.

Designed to cross-correlate templates generated by template_gen function
with data and output the detections.

:copyright:
    EQcorrscan developers.

:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import ast
import copy
import os
import shutil
import logging

from obspy import UTCDateTime, Stream
from obspy.core.event import (
    StationMagnitude, Magnitude, ResourceIdentifier, WaveformStreamID,
    CreationInfo, StationMagnitudeContribution)

from eqcorrscan.core.match_filter.matched_filter import _group_process
from eqcorrscan.core.match_filter.detection import Detection, get_catalog
from eqcorrscan.utils.plotting import cumulative_detections
from eqcorrscan.utils.mag_calc import relative_magnitude

Logger = logging.getLogger(__name__)


class Family(object):
    """
    Container for Detection objects from a single template.

    :type template: eqcorrscan.core.match_filter.Template
    :param template: The template used to detect the family
    :type detections: list
    :param detections: list of Detection objects
    :type catalog: obspy.core.event.Catalog
    :param catalog:
        Catalog of detections, with information for the individual detections.
    """

    def __init__(self, template, detections=None, catalog=None):
        """Instantiation of Family object."""
        self.template = template
        if isinstance(detections, Detection):
            detections = [detections]
        self.detections = detections or []
        self.__catalog = get_catalog(self.detections)
        if catalog:
            Logger.warning("Setting catalog directly is no-longer supported, "
                           "now generated from detections.")

    @property
    def catalog(self):
        if len(self.__catalog) != len(self.detections):
            self.__catalog = get_catalog(self.detections)
        return self.__catalog

    @catalog.setter
    def catalog(self, catalog):
        raise NotImplementedError(
            "Setting catalog directly is no-longer supported")

    def __repr__(self):
        """
        Print method on Family.

        :return: str

        .. rubric:: Example

        >>> from eqcorrscan import Template
        >>> family = Family(template=Template(name='a'))
        >>> print(family)
        Family of 0 detections from template a
        """
        print_str = ('Family of %s detections from template %s' %
                     (len(self.detections), self.template.name))
        return print_str

    def __add__(self, other):
        """
        Extend method. Used for '+'

        .. rubric:: Example

        >>> from eqcorrscan import Template, Detection
        >>> family_a = Family(template=Template(name='a'))
        >>> family_b = Family(template=Template(name='a'))
        >>> family_c = family_a + family_b
        >>> print(family_c)
        Family of 0 detections from template a


        Can only extend family with the family of detections from the same
        template:

        >>> family_a = Family(template=Template(name='a'))
        >>> family_b = Family(template=Template(name='b'))
        >>> family_c = family_a + family_b # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        NotImplementedError: Templates do not match


        Can extend by adding a detection from the same template.

        >>> family_a = Family(template=Template(name='a'))
        >>> detection = Detection(
        ...     template_name='a', detect_time=UTCDateTime(), no_chans=5,
        ...     detect_val=2.5, threshold=1.23, typeofdet='corr',
        ...     threshold_type='MAD', threshold_input=8.0)
        >>> family = family_a + detection
        >>> print(family)
        Family of 1 detections from template a


        Will not work if detections are made using a different Template.

        >>> family_a = Family(template=Template(name='a'))
        >>> detection = Detection(
        ...     template_name='b', detect_time=UTCDateTime(), no_chans=5,
        ...     detect_val=2.5, threshold=1.23, typeofdet='corr',
        ...     threshold_type='MAD', threshold_input=8.0)
        >>> family = family_a + detection # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        NotImplementedError: Templates do not match


        Cannot extent a family with a list, or another object.

        >>> family_a = Family(template=Template(name='a'))
        >>> family = family_a + ['albert'] # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        NotImplementedError: Can only extend with a Detection of Family object.
        """
        return self.copy().__iadd__(other)

    def __iadd__(self, other):
        """
        Rich method '+='

        .. rubric:: Example

        >>> from eqcorrscan import Template
        >>> family_a = Family(template=Template(name='a'))
        >>> family_b = Family(template=Template(name='a'))
        >>> family_a += family_b
        >>> print(family_a)
        Family of 0 detections from template a
        """
        if isinstance(other, Family):
            if other.template == self.template:
                self.detections.extend(other.detections)
                self.__catalog.events.extend(get_catalog(other.detections))
            else:
                raise NotImplementedError('Templates do not match')
        elif isinstance(other, Detection) and other.template_name \
                == self.template.name:
            self.detections.append(other)
            self.__catalog.events.extend(get_catalog([other]))
        elif isinstance(other, Detection):
            raise NotImplementedError('Templates do not match')
        else:
            raise NotImplementedError('Can only extend with a Detection or '
                                      'Family object.')
        return self

    def __eq__(self, other, verbose=False):
        """
        Check equality, rich comparison operator '=='

        .. rubric:: Example

        >>> from eqcorrscan import Template, Detection
        >>> family_a = Family(template=Template(name='a'), detections=[])
        >>> family_b = Family(template=Template(name='a'), detections=[])
        >>> family_a == family_b
        True
        >>> family_c = Family(template=Template(name='b'))
        >>> family_c == family_a
        False


        Test if families are equal without the same detections

        >>> family_a = Family(
        ...     template=Template(name='a'), detections=[
        ...     Detection(template_name='a', detect_time=UTCDateTime(0),
        ...               no_chans=8, detect_val=4.2, threshold=1.2,
        ...               typeofdet='corr', threshold_type='MAD',
        ...               threshold_input=8.0)])
        >>> family_b = Family(
        ...     template=Template(name='a'), detections=[
        ...     Detection(template_name='a', detect_time=UTCDateTime(0) + 10,
        ...               no_chans=8, detect_val=4.2, threshold=1.2,
        ...               typeofdet='corr', threshold_type='MAD',
        ...               threshold_input=8.0)])
        >>> family_a == family_b
        False
        >>> family_c = Family(
        ...     template=Template(name='a'), detections=[
        ...     Detection(template_name='a', detect_time=UTCDateTime(0),
        ...               no_chans=8, detect_val=4.2, threshold=1.2,
        ...               typeofdet='corr', threshold_type='MAD',
        ...               threshold_input=8.0),
        ...     Detection(template_name='a', detect_time=UTCDateTime(0) + 10,
        ...               no_chans=8, detect_val=4.5, threshold=1.2,
        ...               typeofdet='corr', threshold_type='MAD',
        ...               threshold_input=8.0)])
        >>> family_a == family_c
        False
        """
        if not self.template.__eq__(other.template, verbose=verbose):
            return False
        if len(self.detections) != len(other.detections):
            return False
        if len(self.detections) != 0 and len(other.detections) != 0:
            for det, other_det in zip(self.sort().detections,
                                      other.sort().detections):
                if not det.__eq__(other_det, verbose=verbose):
                    return False
        # currently not checking for catalog...
        if len(self.catalog) != len(other.catalog):
            return False
        return True

    def __ne__(self, other):
        """
        Rich comparison operator '!='

        .. rubric:: Example

        >>> from eqcorrscan import Template, Detection
        >>> family_a = Family(
        ...     template=Template(name='a'), detections=[
        ...     Detection(template_name='a', detect_time=UTCDateTime(0),
        ...               no_chans=8, detect_val=4.2, threshold=1.2,
        ...               typeofdet='corr', threshold_type='MAD',
        ...               threshold_input=8.0)])
        >>> family_b = Family(
        ...     template=Template(name='a'), detections=[
        ...     Detection(template_name='a', detect_time=UTCDateTime(),
        ...               no_chans=8, detect_val=4.2, threshold=1.2,
        ...               typeofdet='corr', threshold_type='MAD',
        ...               threshold_input=8.0)])
        >>> family_a != family_b
        True
        """
        return not self.__eq__(other)

    def __getitem__(self, index):
        """
        Retrieve a detection or series of detections from the Family.

        .. rubric:: Example

        >>> from eqcorrscan import Template, Detection
        >>> family = Family(
        ...     template=Template(name='a'), detections=[
        ...     Detection(template_name='a', detect_time=UTCDateTime(0),
        ...               no_chans=8, detect_val=4.2, threshold=1.2,
        ...               typeofdet='corr', threshold_type='MAD',
        ...               threshold_input=8.0),
        ...     Detection(template_name='a', detect_time=UTCDateTime(0) + 10,
        ...               no_chans=8, detect_val=4.5, threshold=1.2,
        ...               typeofdet='corr', threshold_type='MAD',
        ...               threshold_input=8.0)])
        >>> isinstance(family[0], Detection)
        True
        >>> len(family[0:])
        2
        """
        return self.detections.__getitem__(index)

    def __len__(self):
        """Number of detections in Family.

        .. rubric:: Example

        >>> from eqcorrscan import Template, Detection
        >>> family = Family(
        ...     template=Template(name='a'), detections=[
        ...     Detection(template_name='a', detect_time=UTCDateTime(0),
        ...               no_chans=8, detect_val=4.2, threshold=1.2,
        ...               typeofdet='corr', threshold_type='MAD',
        ...               threshold_input=8.0),
        ...     Detection(template_name='a', detect_time=UTCDateTime(0) + 10,
        ...               no_chans=8, detect_val=4.5, threshold=1.2,
        ...               typeofdet='corr', threshold_type='MAD',
        ...               threshold_input=8.0)])
        >>> print(len(family))
        2
        """
        return len(self.detections)

    def _uniq(self):
        """
        Get list of unique detections.
        Works in place.

        .. rubric:: Example

        >>> from eqcorrscan import Template, Detection
        >>> family = Family(
        ...     template=Template(name='a'), detections=[
        ...     Detection(template_name='a', detect_time=UTCDateTime(0),
        ...               no_chans=8, detect_val=4.2, threshold=1.2,
        ...               typeofdet='corr', threshold_type='MAD',
        ...               threshold_input=8.0),
        ...     Detection(template_name='a', detect_time=UTCDateTime(0) + 10,
        ...               no_chans=8, detect_val=4.5, threshold=1.2,
        ...               typeofdet='corr', threshold_type='MAD',
        ...               threshold_input=8.0),
        ...     Detection(template_name='a', detect_time=UTCDateTime(0) + 10,
        ...               no_chans=8, detect_val=4.5, threshold=1.2,
        ...               typeofdet='corr', threshold_type='MAD',
        ...               threshold_input=8.0)])
        >>> len(family)
        3
        >>> len(family._uniq())
        2
        """
        _detections = []
        [_detections.append(d) for d in self.detections
         if not _detections.count(d)]
        self.detections = _detections
        return self

    def sort(self):
        """Sort by detection time.

        .. rubric:: Example

        >>> from eqcorrscan import Template, Detection
        >>> family = Family(
        ...     template=Template(name='a'), detections=[
        ...     Detection(template_name='a', detect_time=UTCDateTime(0) + 200,
        ...               no_chans=8, detect_val=4.2, threshold=1.2,
        ...               typeofdet='corr', threshold_type='MAD',
        ...               threshold_input=8.0),
        ...     Detection(template_name='a', detect_time=UTCDateTime(0),
        ...               no_chans=8, detect_val=4.5, threshold=1.2,
        ...               typeofdet='corr', threshold_type='MAD',
        ...               threshold_input=8.0),
        ...     Detection(template_name='a', detect_time=UTCDateTime(0) + 10,
        ...               no_chans=8, detect_val=4.5, threshold=1.2,
        ...               typeofdet='corr', threshold_type='MAD',
        ...               threshold_input=8.0)])
        >>> family[0].detect_time
        UTCDateTime(1970, 1, 1, 0, 3, 20)
        >>> family.sort()[0].detect_time
        UTCDateTime(1970, 1, 1, 0, 0)
        """
        self.detections = sorted(self.detections, key=lambda d: d.detect_time)
        return self

    def copy(self):
        """
        Returns a copy of the family.

        :return: Copy of family

        .. rubric:: Example

        >>> from eqcorrscan import Template, Detection
        >>> family = Family(
        ...     template=Template(name='a'), detections=[
        ...     Detection(template_name='a', detect_time=UTCDateTime(0),
        ...               no_chans=8, detect_val=4.2, threshold=1.2,
        ...               typeofdet='corr', threshold_type='MAD',
        ...               threshold_input=8.0),
        ...     Detection(template_name='a', detect_time=UTCDateTime(0) + 10,
        ...               no_chans=8, detect_val=4.5, threshold=1.2,
        ...               typeofdet='corr', threshold_type='MAD',
        ...               threshold_input=8.0)])
        >>> family == family.copy()
        True
        """
        return copy.deepcopy(self)

    def append(self, other):
        """
        Add another family or detection to the family.

        .. rubric:: Example

        Append a family to a family

        >>> from eqcorrscan import Template, Detection
        >>> family_a = Family(template=Template(name='a'))
        >>> family_b = Family(template=Template(name='a'))
        >>> family_a.append(family_b)
        Family of 0 detections from template a


        Append a detection to the family

        >>> family_a = Family(template=Template(name='a'))
        >>> detection = Detection(
        ...     template_name='a', detect_time=UTCDateTime(), no_chans=5,
        ...     detect_val=2.5, threshold=1.23, typeofdet='corr',
        ...     threshold_type='MAD', threshold_input=8.0)
        >>> family_a.append(detection)
        Family of 1 detections from template a
        """
        return self.__add__(other)

    def plot(self, plot_grouped=False):
        """
        Plot the cumulative number of detections in time.

        .. rubric:: Example

        >>> from eqcorrscan import Template, Detection
        >>> family = Family(
        ...     template=Template(name='a'), detections=[
        ...     Detection(template_name='a', detect_time=UTCDateTime(0) + 200,
        ...               no_chans=8, detect_val=4.2, threshold=1.2,
        ...               typeofdet='corr', threshold_type='MAD',
        ...               threshold_input=8.0),
        ...     Detection(template_name='a', detect_time=UTCDateTime(0),
        ...               no_chans=8, detect_val=4.5, threshold=1.2,
        ...               typeofdet='corr', threshold_type='MAD',
        ...               threshold_input=8.0),
        ...     Detection(template_name='a', detect_time=UTCDateTime(0) + 10,
        ...               no_chans=8, detect_val=4.5, threshold=1.2,
        ...               typeofdet='corr', threshold_type='MAD',
        ...               threshold_input=8.0)])
        >>> family.plot(plot_grouped=True)  # doctest: +SKIP

        .. plot::

            from eqcorrscan.core.match_filter import Family, Template
            from eqcorrscan.core.match_filter import Detection
            from obspy import UTCDateTime
            family = Family(
                template=Template(name='a'), detections=[
                Detection(template_name='a', detect_time=UTCDateTime(0) + 200,
                          no_chans=8, detect_val=4.2, threshold=1.2,
                          typeofdet='corr', threshold_type='MAD',
                          threshold_input=8.0),
                Detection(template_name='a', detect_time=UTCDateTime(0),
                          no_chans=8, detect_val=4.5, threshold=1.2,
                          typeofdet='corr', threshold_type='MAD',
                          threshold_input=8.0),
                Detection(template_name='a', detect_time=UTCDateTime(0) + 10,
                          no_chans=8, detect_val=4.5, threshold=1.2,
                          typeofdet='corr', threshold_type='MAD',
                          threshold_input=8.0)])
            family.plot(plot_grouped=True)
        """
        cumulative_detections(
            detections=self.detections, plot_grouped=plot_grouped)

    def write(self, filename, format='tar'):
        """
        Write Family out, select output format.

        :type format: str
        :param format:
            One of either 'tar', 'csv', or any obspy supported
            catalog output.
        :type filename: str
        :param filename: Path to write file to.

        .. Note:: csv format will write out detection objects, all other
            outputs will write the catalog.  These cannot be rebuilt into
            a Family object.  The only format that can be read back into
            Family objects is the 'tar' type.

        .. Note:: csv format will append detections to filename, all others
            will overwrite any existing files.

        .. rubric:: Example

        >>> from eqcorrscan import Template, Detection
        >>> from obspy import read
        >>> family = Family(
        ...     template=Template(name='a', st=read()), detections=[
        ...     Detection(template_name='a', detect_time=UTCDateTime(0) + 200,
        ...               no_chans=8, detect_val=4.2, threshold=1.2,
        ...               typeofdet='corr', threshold_type='MAD',
        ...               threshold_input=8.0),
        ...     Detection(template_name='a', detect_time=UTCDateTime(0),
        ...               no_chans=8, detect_val=4.5, threshold=1.2,
        ...               typeofdet='corr', threshold_type='MAD',
        ...               threshold_input=8.0),
        ...     Detection(template_name='a', detect_time=UTCDateTime(0) + 10,
        ...               no_chans=8, detect_val=4.5, threshold=1.2,
        ...               typeofdet='corr', threshold_type='MAD',
        ...               threshold_input=8.0)])
        >>> family.write('test_family')
        """
        from eqcorrscan.core.match_filter.party import Party

        Party(families=[self]).write(filename=filename, format=format)
        return

    def lag_calc(self, stream, pre_processed, shift_len=0.2, min_cc=0.4,
                 horizontal_chans=['E', 'N', '1', '2'], vertical_chans=['Z'],
                 cores=1, interpolate=False, plot=False, plotdir=None,
                 parallel=True, process_cores=None, ignore_bad_data=False,
                 relative_magnitudes=False, **kwargs):
        """
        Compute picks based on cross-correlation alignment.

        Works in place on events in the Family

        :type stream: obspy.core.stream.Stream
        :param stream:
            All the data needed to cut from - can be a gappy Stream.
        :type pre_processed: bool
        :param pre_processed:
            Whether the stream has been pre-processed or not to match the
            templates. See note below.
        :type shift_len: float
        :param shift_len:
            Shift length allowed for the pick in seconds, will be
            plus/minus this amount - default=0.2
        :type min_cc: float
        :param min_cc:
            Minimum cross-correlation value to be considered a pick,
            default=0.4.
        :type horizontal_chans: list
        :param horizontal_chans:
            List of channel endings for horizontal-channels, on which
            S-picks will be made.
        :type vertical_chans: list
        :param vertical_chans:
            List of channel endings for vertical-channels, on which P-picks
            will be made.
        :type cores: int
        :param cores:
            Number of cores to use in parallel processing, defaults to one.
        :type interpolate: bool
        :param interpolate:
            Interpolate the correlation function to achieve sub-sample
            precision.
        :type plot: bool
        :param plot:
            To generate a plot for every detection or not, defaults to False.
        :type plotdir: str
        :param plotdir:
            The path to save plots to. If `plotdir=None` (default) then the
            figure will be shown on screen.
        :type parallel: bool
        :param parallel: Turn parallel processing on or off.
        :type process_cores: int
        :param process_cores:
            Number of processes to use for pre-processing (if different to
            `cores`).
        :type ignore_bad_data: bool
        :param ignore_bad_data:
            If False (default), errors will be raised if data are excessively
            gappy or are mostly zeros. If True then no error will be raised,
            but an empty trace will be returned (and not used in detection).
        :type relative_magnitudes: bool
        :param relative_magnitudes:
            Whether to calculate relative magnitudes or not. See
            :func:`eqcorrscan.utils.mag_calc.relative_magnitude` for more
            information. Keyword arguments `noise_window`, `signal_window` and
            `min_snr` can be passed as additional keyword arguments to pass
            through to `eqcorrscan.utils.mag_calc.relative_magnitude`.

        :returns:
            Catalog of events with picks.  No origin information is included.
            These events can then be written out via
            :func:`obspy.core.event.Catalog.write`, or to Nordic Sfiles using
            :func:`eqcorrscan.utils.sfile_util.eventtosfile` and located
            externally.
        :rtype: obspy.core.event.Catalog

        .. Note::
            Note on pre-processing: You can provide a pre-processed stream,
            which may be beneficial for detections over large time periods
            (the stream can have gaps, which reduces memory usage).  However,
            in this case the processing steps are not checked, so you must
            ensure that the template in the Family has the same sampling
            rate and filtering as the stream.
            If pre-processing has not be done then the data will be processed
            according to the parameters in the template.

        .. Note::
            Picks are corrected for the template pre-pick time.
        """
        from eqcorrscan.core.lag_calc import xcorr_pick_family

        processed_stream = self._process_streams(
            stream=stream, pre_processed=pre_processed,
            process_cores=process_cores, parallel=parallel,
            ignore_bad_data=ignore_bad_data)
        picked_dict = xcorr_pick_family(
            family=self, stream=processed_stream, shift_len=shift_len,
            min_cc=min_cc, horizontal_chans=horizontal_chans,
            vertical_chans=vertical_chans, cores=cores,
            interpolate=interpolate, plot=plot, plotdir=plotdir)
        for detection_id, event in picked_dict.items():
            for pick in event.picks:
                pick.time += self.template.prepick
            d = [d for d in self.detections if d.id == detection_id][0]
            d.event.picks = event.picks
        if relative_magnitudes:
            self.relative_magnitudes(
                stream=processed_stream, pre_processed=True, min_cc=min_cc,
                **kwargs)
        return self.catalog

    def relative_magnitudes(self, stream, pre_processed, process_cores=1,
                            ignore_bad_data=False, parallel=False, min_cc=0.4,
                            **kwargs):
        """
        Compute relative magnitudes for the detections.

        Works in place on events in the Family

        :type stream: obspy.core.stream.Stream
        :param stream:
            All the data needed to cut from - can be a gappy Stream.
        :type pre_processed: bool
        :param pre_processed:
            Whether the stream has been pre-processed or not to match the
            templates. See note below.
        :param parallel: Turn parallel processing on or off.
        :type process_cores: int
        :param process_cores:
            Number of processes to use for pre-processing (if different to
            `cores`).
        :type ignore_bad_data: bool
        :param ignore_bad_data:
            If False (default), errors will be raised if data are excessively
            gappy or are mostly zeros. If True then no error will be raised,
            but an empty trace will be returned (and not used in detection).
        :type min_cc: float
        :param min_cc: Minimum correlation for magnitude to be computed.
        :param kwargs:
            Keyword arguments passed to `utils.mag_calc.relative_mags`

        .. Note::
            Note on pre-processing: You can provide a pre-processed stream,
            which may be beneficial for detections over large time periods
            (the stream can have gaps, which reduces memory usage).  However,
            in this case the processing steps are not checked, so you must
            ensure that the template in the Family has the same sampling
            rate and filtering as the stream.
            If pre-processing has not be done then the data will be processed
            according to the parameters in the template.
        """
        processed_stream = self._process_streams(
            stream=stream, pre_processed=pre_processed,
            process_cores=process_cores, parallel=parallel,
            ignore_bad_data=ignore_bad_data)
        for detection in self.detections:
            event = detection.event
            if event is None:
                continue
            corr_dict = {}
            for pick in event.picks:
                if len(pick.comments) == 0:
                    continue
                try:
                    corr = float(pick.comments[0].text.split("=")[-1])
                except IndexError:
                    continue
                corr_dict.update({pick.waveform_id.get_seed_string(): corr})
            if len(corr_dict) == 0:
                Logger.info("Correlations not run for {0}".format(
                    detection.id))
                continue
            template = self.template
            try:
                t_mag = (
                        template.event.preferred_magnitude() or
                        template.event.magnitudes[0])
            except IndexError:
                Logger.info(
                    "No template magnitude, relative magnitudes cannot"
                    " be computed for {0}".format(event.resource_id))
                continue
            # Set the signal-window to be the template length
            signal_window = (-template.prepick,
                             min([tr.stats.npts * tr.stats.delta
                                  for tr in template.st]) - template.prepick)
            delta_mag = relative_magnitude(
                st1=template.st, st2=processed_stream,
                event1=template.event, event2=event,
                correlations=corr_dict, min_cc=min_cc,
                signal_window=signal_window, **kwargs)
            # Add station magnitudes
            sta_contrib = []
            av_mag = 0.0
            for seed_id, _delta_mag in delta_mag.items():
                sta_mag = StationMagnitude(
                    mag=t_mag.mag + _delta_mag,
                    magnitude_type=t_mag.magnitude_type,
                    method_id=ResourceIdentifier("relative"),
                    waveform_id=WaveformStreamID(seed_string=seed_id),
                    creation_info=CreationInfo(
                        author="EQcorrscan",
                        creation_time=UTCDateTime()))
                event.station_magnitudes.append(sta_mag)
                sta_contrib.append(StationMagnitudeContribution(
                    station_magnitude_id=sta_mag.resource_id,
                    weight=1.))
                av_mag += sta_mag.mag
            if len(delta_mag) > 0:
                av_mag /= len(delta_mag)
                # Compute average magnitude
                event.magnitudes.append(Magnitude(
                    mag=av_mag, magnitude_type=t_mag.magnitude_type,
                    method_id=ResourceIdentifier("relative"),
                    station_count=len(delta_mag),
                    evaluation_mode="manual",
                    station_magnitude_contributions=sta_contrib,
                    creation_info=CreationInfo(
                        author="EQcorrscan",
                        creation_time=UTCDateTime())))
        return self.catalog

    def _process_streams(self, stream, pre_processed, process_cores=1,
                         parallel=False, ignore_bad_data=False,
                         select_used_chans=True):
        """
        Process a stream based on the template parameters.
        """
        if select_used_chans:
            template_stream = Stream()
            for tr in self.template.st:
                template_stream += stream.select(
                    network=tr.stats.network, station=tr.stats.station,
                    location=tr.stats.location, channel=tr.stats.channel)
        else:
            template_stream = stream
        if not pre_processed:
            processed_streams = _group_process(
                template_group=[self.template], cores=process_cores,
                parallel=parallel, stream=template_stream.copy(),
                daylong=False, ignore_length=False, overlap=0.0,
                ignore_bad_data=ignore_bad_data)
            processed_stream = Stream()
            for p in processed_streams:
                processed_stream += p
            processed_stream.merge(method=1)
            Logger.debug(processed_stream)
        else:
            processed_stream = stream.merge()
        return processed_stream

    def extract_streams(self, stream, length, prepick):
        """
        Generate a dictionary of cut streams around detections.

        :type stream: `obspy.core.stream.Stream`
        :param stream: Stream of data to cut from
        :type length: float
        :param length: Length of data to extract in seconds.
        :type prepick: float
        :param prepick:
            Length before the expected pick on each channel to start the cut
            stream in seconds.

        :rtype: dict
        :returns: Dictionary of cut streams keyed by detection id.
        """
        return {d.id: d.extract_stream(
            stream=stream, length=length, prepick=prepick)
            for d in self.detections}


def _write_family(family, filename):
    """
    Write a family to a csv file.

    :type family: :class:`eqcorrscan.core.match_filter.Family`
    :param family: Family to write to file
    :type filename: str
    :param filename: File to write to.
    """
    with open(filename, 'w') as f:
        for detection in family.detections:
            det_str = ''
            for key in detection.__dict__.keys():
                if key == 'event' and detection.__dict__[key] is not None:
                    value = str(detection.event.resource_id)
                elif key in ['threshold', 'detect_val', 'threshold_input']:
                    value = format(detection.__dict__[key], '.32f').rstrip('0')
                else:
                    value = str(detection.__dict__[key])
                det_str += key + ': ' + value + '; '
            f.write(det_str + '\n')
    return


def _read_family(fname, all_cat, template, encoding="UTF8",
                 estimate_origin=True):
    """
    Internal function to read csv family files.

    :type fname: str
    :param fname: Filename
    :return: list of Detection
    """
    detections = []
    with open(fname, 'rb') as _f:
        lines = _f.read().decode(encoding).splitlines()
    for line in lines:
        det_dict = {}
        gen_event = False
        for key_pair in line.rstrip().split(';'):
            key = key_pair.split(': ')[0].strip()
            value = key_pair.split(': ')[-1].strip()
            if key == 'event':
                if len(all_cat) == 0:
                    gen_event = True
                    continue
                el = [e for e in all_cat
                      if str(e.resource_id).split('/')[-1] == value][0]
                det_dict.update({'event': el})
            elif key == 'detect_time':
                det_dict.update(
                    {'detect_time': UTCDateTime(value)})
            elif key == 'chans':
                det_dict.update({'chans': ast.literal_eval(value)})
            elif key in ['template_name', 'typeofdet', 'id',
                         'threshold_type']:
                det_dict.update({key: value})
            elif key == 'no_chans':
                det_dict.update({key: int(float(value))})
            elif len(key) == 0:
                continue
            else:
                det_dict.update({key: float(value)})
        detection = Detection(**det_dict)
        if gen_event:
            detection._calculate_event(
                template=template, estimate_origin=estimate_origin,
                correct_prepick=True)
        detections.append(detection)
    return detections


if __name__ == "__main__":
    import doctest

    doctest.testmod()
    # List files to be removed after doctest
    cleanup = ['test_family.tgz']
    for f in cleanup:
        if os.path.isfile(f):
            os.remove(f)
        elif os.path.isdir(f):
            shutil.rmtree(f)
